# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to describe classes for the calculation of wakes and impedances to
be used either alone or by InducedVoltage objects described in impedance.py.
The module consists of a parent class called _ImpedanceObject and several child
classes, as for example InputTable, Resonators and TravelingWaveCavity.**

:Authors: **Danilo Quartullo**, **Alexandre Lasheen**, **Juan F. Esteban Mueller**,
    **Markus Schwarz**
'''

from __future__ import division, print_function

from builtins import range

import mpmath
import numpy as np
from scipy import integrate
from scipy.constants import c, physical_constants
from scipy.special import airy
from scipy.special import gamma as gamma_func
from scipy.special import kv, polygamma

from ..utils import bmath as bm


class _ImpedanceObject:

    r"""
    Parent impedance object to implement required methods and attributes
    common to all the child classes. The attributes are initialised to 0 but
    they are overwritten by float arrays when the child classes are used.
    """

    def __init__(self):
        # Time array of the wake in s
        self.time_array = 0

        # Wake array in :math:`\Omega / s`
        self.wake = 0

        # Frequency array of the impedance in Hz
        self.frequency_array = 0

        # Impedance array in :math:`\Omega`
        self.impedance = 0

    def wake_calc(self, *args):
        r"""
        Method required to compute the wake function. Returns an error if
        called from an object which does not implement this method.
        """
        # WrongCalcError
        raise NotImplementedError('wake_calc() method not implemented in this class. ' +
                                  'This object is probably meant to be used in the ' +
                                  'frequency domain')

    def imped_calc(self, *args):
        r"""
        Method required to compute the impedance. Returns an error if called
        from an object which does not implement this method.
        """
        # WrongCalcError
        raise NotImplementedError('imped_calc() method not implemented in this class. ' +
                                  'This object is probably meant to be used in the ' +
                                  'time domain')


class InputTable(_ImpedanceObject):
    r"""
    Intensity effects from impedance and wake tables.
    If the constructor takes just two arguments, then a wake table is passed;
    if it takes three arguments, then an impedance table is passed. Be careful
    that if you input a wake, the input wake for W(t=0) should be already
    divided by two (beam loading theorem) ; and that if you input impedance,
    only the positive  frequencies of the impedance is needed (the impedance
    will be assumed to be Hermitian (Real part symmetric and Imaginary part
    antisymmetric).Note that we add the point (f, Z(f)) = (0, 0) to the
    frequency and impedance arrays derived from the table.

    Parameters
    ----------
    input_1 : float array
        Time array of the wake in s or frequency array of the impedance in Hz
    input_2 : float array
        Wake array in :math:`\Omega / s` or real part of impedance
        in :math:`\Omega`
    input_3 : float array
        Imaginary part of impedance in :math:`\Omega`


    Attributes
    ----------
    time_array : float array
        Input time array of the wake in s
    wake_array : float array
        Input wake array in :math:`\Omega / s`
    frequency_array_loaded : float array
        Input frequency array of the impedance in Hz
    Re_Z_array_loaded : float array
        Input real part of impedance in :math:`\Omega`
    Im_Z_array_loaded : float array
        Input imaginary part of impedance in :math:`\Omega`
    impedance_loaded : complex array
        Input impedance array in :math:`\Omega + j \Omega`

    Examples
    --------
    >>> ##### WAKEFIELD TABLE
    >>> time = np.array([1.1,2.2,3.3])
    >>> wake = np.array([4.4,5.5,6.6])
    >>> table = InputTable(time, wake)
    >>> new_time_array = np.array([7.7,8.8,9.9])
    >>> table.wake_calc(new_time_array)
    >>>
    >>> ##### IMPEDANCE TABLE
    >>> frequency = np.array([1.1,2.2,3.3])
    >>> real_part = np.array([4.4,5.5,6.6])
    >>> imaginary_part = np.array([7.7,8.8,9.9])
    >>> table2 = InputTable(frequency, real_part, imaginary_part)
    >>> new_freq_array = np.array([7.7,8.8,9.9])
    >>> table2.imped_calc(new_freq_array)

    """

    def __init__(self, input_1, input_2, input_3=None):

        _ImpedanceObject.__init__(self)

        if input_3 is None:
            # Time array of the wake in s
            self.time_array = input_1
            # Wake array in :math:`\Omega / s`
            self.wake_array = input_2
        else:
            # Frequency array of the impedance in Hz
            self.frequency_array_loaded = input_1
            # Real part of impedance in :math:`\Omega`
            self.Re_Z_array_loaded = input_2
            # Imaginary part of impedance in :math:`\Omega`
            self.Im_Z_array_loaded = input_3
            # Impedance array in :math:`\Omega`
            self.impedance_loaded = (self.Re_Z_array_loaded + 1j *
                                     self.Im_Z_array_loaded)

            if self.frequency_array_loaded[0] != 0:
                self.frequency_array_loaded = np.hstack((0,
                                                         self.frequency_array_loaded))
                self.Re_Z_array_loaded = np.hstack((0, self.Re_Z_array_loaded))
                self.Im_Z_array_loaded = np.hstack((0, self.Im_Z_array_loaded))

    def wake_calc(self, new_time_array):
        r"""
        The wake from the table is interpolated using the new time array.

        Parameters
        ----------
        new_time_array : float array
            Input time array in s

        Attributes
        ----------
        new_time_array : float array
            Input time array in s
        wake: float array
            Output interpolated wake in :math:`\Omega / s`
        """

        self.new_time_array = new_time_array
        self.wake = np.interp(self.new_time_array, self.time_array,
                              self.wake_array, right=0)

    def imped_calc(self, new_frequency_array):
        r"""
        The impedance from the table is interpolated using the new frequency
        array.

        Parameters
        ----------
        new_frequency_array : float array
            frequency array in :math:`\Omega`

        Attributes
        ----------
        frequency_array : float array
            frequency array in :math:`\Omega`
        Re_Z_array : float array
            Real part of the interpolated impedance in :math:`\Omega`
        Im_Z_array : float array
            Imaginary part of the interpolated impedance in :math:`\Omega`
        impedance : complex array
            Output interpolated impedance array in :math:`\Omega + j \Omega`
        """

        Re_Z = np.interp(new_frequency_array, self.frequency_array_loaded,
                         self.Re_Z_array_loaded, right=0)
        Im_Z = np.interp(new_frequency_array, self.frequency_array_loaded,
                         self.Im_Z_array_loaded, right=0)
        self.frequency_array = new_frequency_array
        self.Re_Z_array = Re_Z
        self.Im_Z_array = Im_Z
        self.impedance = Re_Z + 1j * Im_Z


class Resonators(_ImpedanceObject):
    r"""
    Impedance contribution from resonators, analytic formulas for both wake
    and impedance. The resonant modes (and the corresponding R and Q)
    can be inputed as a list in case of several modes.
    The model is the following:

    .. math::

        Z(f) = \frac{R}{1 + j Q \left(\frac{f}{f_r}-\frac{f_r}{f}\right)}

    .. math::

        W(t>0) = 2\alpha R e^{-\alpha t}\left(\cos{\bar{\omega}t} - \frac{\alpha}{\bar{\omega}}\sin{\bar{\omega}t}\right)

        W(0) = \alpha R

    .. math::

        \omega_r = 2 \pi f_r

        \alpha = \frac{\omega_r}{2Q}

        \bar{\omega} = \sqrt{\omega_r^2 - \alpha^2}

    Parameters
    ----------
    R_S : float list
        Shunt impepdance in :math:`\Omega`
    frequency_R : float list
        Resonant frequency in Hz
    Q : float list
        Quality factor
    method: string
        It defines which algorithm to use to calculate the impedance (C++ or
        Python)

    Attributes
    ----------
    R_S : float array
        Shunt impepdance in :math:`\Omega`
    frequency_R : float array
        Resonant frequency in Hz
    Q : float array
        Quality factor
    n_resonators : int
        number of resonators

    Examples
    ----------
    >>> R_S = [1, 2, 3]
    >>> frequency_R = [1, 2, 3]
    >>> Q = [1, 2, 3]
    >>> resonators = Resonators(R_S, frequency_R, Q)
    >>> time = np.array(1,2,3)
    >>> resonators.wake_calc(time)
    >>> frequency = np.array(1,2,3)
    >>> resonators.imped_calc(frequency)
    """

    def __init__(self, R_S, frequency_R, Q, method='c++'):

        _ImpedanceObject.__init__(self)

        # Shunt impepdance in :math:`\Omega`
        self.R_S = np.array([R_S], dtype=float).flatten()

        # Resonant frequency in Hz
        self.frequency_R = np.array([frequency_R], dtype=float).flatten()

        # Quality factor
        self.Q = np.array([Q], dtype=float).flatten()

        # Test if one or more quality factors is smaller than 0.5.
        if np.count_nonzero(self.Q < 0.5) > 0:
            # ResonatorError
            raise RuntimeError('All quality factors Q must be greater or equal 0.5')

        # Number of resonant modes
        self.n_resonators = len(self.R_S)

        if method == 'c++':
            self.imped_calc = self._imped_calc_cpp
        elif method == 'python':
            self.imped_calc = self._imped_calc_python
        else:
            # WrongCalcError
            raise RuntimeError(
                'method for impedance calculation in Resonator object not recognized')

    @property
    def frequency_R(self):
        return self.__frequency_R

    @frequency_R.setter
    def frequency_R(self, frequency_R):
        self.__frequency_R = frequency_R
        self.__omega_R = 2 * np.pi * frequency_R

    # Resonant angular frequency in rad/s
    @property
    def omega_R(self):
        return self.__omega_R

    @omega_R.setter
    def omega_R(self, omega_R):
        self.__frequency_R = omega_R / 2 / np.pi
        self.__omega_R = omega_R

    def wake_calc(self, time_array):
        r"""
        Wake calculation method as a function of time.

        Parameters
        ----------
        time_array : float array
            Input time array in s

        Attributes
        ----------
        time_array : float array
            Input time array in s
        wake : float array
            Output wake in :math:`\Omega / s`
        """

        self.time_array = time_array
        self.wake = np.zeros(self.time_array.shape, dtype=bm.precision.real_t, order='C')

        for i in range(0, self.n_resonators):

            alpha = self.omega_R[i] / (2 * self.Q[i])
            omega_bar = np.sqrt(self.omega_R[i] ** 2 - alpha ** 2)

            self.wake += ((np.sign(self.time_array) + 1) * self.R_S[i]
                          * alpha * np.exp(-alpha * self.time_array)
                          * (bm.cos(omega_bar * self.time_array) - alpha /
                             omega_bar * bm.sin(omega_bar * self.time_array)))

    def _imped_calc_python(self, frequency_array):
        r"""
        Impedance calculation method as a function of frequency using Python.

        Parameters
        ----------
        frequency_array : float array
            Input frequency array in Hz

        Attributes
        ----------
        frequency_array : float array
            nput frequency array in Hz
        impedance : complex array
            Output impedance in :math:`\Omega + j \Omega`
        """

        self.frequency_array = frequency_array
        self.impedance = np.zeros(len(self.frequency_array), dtype=bm.precision.complex_t, order='C')

        for i in range(0, self.n_resonators):

            self.impedance[1:] += self.R_S[i] / (1 + 1j * self.Q[i]
                                                 * (self.frequency_array[1:] / self.frequency_R[i] -
                                                    self.frequency_R[i] / self.frequency_array[1:]))

    def _imped_calc_cpp(self, frequency_array):
        r"""
        Impedance calculation method as a function of frequency optimised in C++

        Parameters
        ----------
        frequency_array : float array
            Input frequency array in Hz

        Attributes
        ----------
        frequency_array : float array
            nput frequency array in Hz
        impedance : complex array
            Output impedance in :math:`\Omega + j \Omega`
        """

        self.frequency_array = frequency_array
        self.impedance = bm.fast_resonator(self.R_S, self.Q,
                                           self.frequency_array,
                                           self.frequency_R)


class TravelingWaveCavity(_ImpedanceObject):
    r"""
    Impedance contribution from travelling wave cavities, analytic formulas for
    both wake and impedance. The resonance modes (and the corresponding R and a)
    can be inputed as a list in case of several modes.

    The model is the following:

    .. math::
        Z &= Z_+ + Z_- \\
        Z_-(f) &= R \left[\left(\frac{\sin{\frac{a(f-f_r)}{2}}}{\frac{a(f-f_r)}{2}}\right)^2 - 2i \frac{a(f-f_r) - \sin{a(f-f_r)}}{\left(a(f-f_r)\right)^2}\right] \\
        Z_+(f) &= R \left[\left(\frac{\sin{\frac{a(f+f_r)}{2}}}{\frac{a(f+f_r)}{2}}\right)^2 - 2i \frac{a(f+f_r) - \sin{a(f+f_r)}}{\left(a(f+f_r)\right)^2}\right]

    .. math::
        W(0<t<\tilde{a}) &= \frac{4R}{\tilde{a}}\left(1-\frac{t}{\tilde{a}}\right)\cos{\omega_r t} \\
        W(0) &= \frac{2R}{\tilde{a}}

    .. math::
        a = 2 \pi \tilde{a}

    Parameters
    ----------
    R_S : float list
        Shunt impepdance in :math:`\Omega`
    frequency_R : float list
        Resonant frequency in Hz
    a_factor : float list
        Damping time a in s

    Attributes
    ----------
    R_S : float array
        Shunt impepdance in :math:`\Omega`
    frequency_R : float array
        Resonant frequency in Hz
    a_factor : float array
        Damping time a in s
    n_twc : int
        number of resonant modes

    Examples
    ----------
    >>> R_S = [1, 2, 3]
    >>> frequency_R = [1, 2, 3]
    >>> a_factor = [1, 2, 3]
    >>> twc = TravelingWaveCavity(R_S, frequency_R, a_factor)
    >>> time = np.array(1,2,3)
    >>> twc.wake_calc(time)
    >>> frequency = np.array(1,2,3)
    >>> twc.imped_calc(frequency)
    """

    def __init__(self, R_S, frequency_R, a_factor):

        _ImpedanceObject.__init__(self)

        # Shunt impepdance in :math:`\Omega`
        self.R_S = np.array([R_S], dtype=float).flatten()

        # Resonant frequency in Hz
        self.frequency_R = np.array([frequency_R], dtype=float).flatten()

        # Damping time a in s
        self.a_factor = np.array([a_factor], dtype=float).flatten()

        # Number of resonant modes
        self.n_twc = len(self.R_S)

    def wake_calc(self, time_array):
        r"""
        Wake calculation method as a function of time.

        Parameters
        ----------
        time_array : float array
            Input time array in s

        Attributes
        ----------
        time_array : float array
            Input time array in s
        wake : float array
            Output wake in :math:`\Omega / s`
        """

        self.time_array = time_array
        self.wake = np.zeros(self.time_array.shape, dtype=bm.precision.real_t, order='C')
        self.wake = np.zeros(self.time_array.shape)

        for i in range(0, self.n_twc):
            a_tilde = self.a_factor[i] / (2 * np.pi)
            indexes = self.time_array <= a_tilde
            self.wake[indexes] += ((np.sign(self.time_array[indexes]) + 1) * 2
                                   * self.R_S[i] / a_tilde
                                   * (1 - self.time_array[indexes] / a_tilde)
                                   * bm.cos(2 * np.pi * self.frequency_R[i] *
                                            self.time_array[indexes]))

    def imped_calc(self, frequency_array):
        r"""
        Impedance calculation method as a function of frequency.

        Parameters
        ----------
        frequency_array : float array
            Input frequency array in Hz

        Attributes
        ----------
        frequency_array : float array
            nput frequency array in Hz
        impedance : complex array
            Output impedance in :math:`\Omega + j \Omega`
        """

        self.frequency_array = frequency_array
        self.impedance = np.zeros(len(self.frequency_array), dtype=bm.precision.complex_t, order='C')

        for i in range(0, self.n_twc):

            Zplus = self.R_S[i] * ((bm.sin(self.a_factor[i] / 2 *
                                           (self.frequency_array - self.frequency_R[i])) /
                                    (self.a_factor[i] / 2 * (self.frequency_array -
                                                             self.frequency_R[i])))**2 - 2j * (self.a_factor[i] *
                                                                                               (self.frequency_array - self.frequency_R[i]) -
                                                                                               bm.sin(self.a_factor[i] * (self.frequency_array -
                                                                                                                          self.frequency_R[i]))) / (self.a_factor[i] *
                                                                                                                                                    (self.frequency_array - self.frequency_R[i]))**2)

            Zminus = self.R_S[i] * ((bm.sin(self.a_factor[i] / 2 *
                                            (self.frequency_array + self.frequency_R[i])) /
                                     (self.a_factor[i] / 2 * (self.frequency_array
                                                              + self.frequency_R[i])))**2 - 2j * (self.a_factor[i]
                                                                                                  * (self.frequency_array + self.frequency_R[i])
                                                                                                  - bm.sin(self.a_factor[i] * (self.frequency_array
                                                                                                                               + self.frequency_R[i]))) / (self.a_factor[i] *
                                                                                                                                                           (self.frequency_array + self.frequency_R[i]))**2)

            self.impedance += Zplus + Zminus


class ResistiveWall(_ImpedanceObject):
    r"""
    Impedance contribution from resistive wall for a cilindrical beam pipe

    The model is the following:

    .. math::

        Z(f) = \frac{Z_0 c L}{ \pi } \frac{ 1 }{ \left[1 - i \text{sign} f \right] 2  b  c
                                    \sqrt{ \frac{\sigma_c Z_0 c }{ 4 \pi |f| } + i 2 \pi b^2 f } }

    Parameters
    ----------
    pipe_radius : float
        Beam pipe radius in m
    pipe_length : float
        Beam pipe length in m
    resistivity : float
        Beam pipe resistivity in :math:`m / s`
    conductivity : float
        Beam pipe conductivity in :math:`s / m`

    Attributes
    ----------
    pipe_radius : float
        Beam pipe radius in m
    pipe_length : float
        Beam pipe length in m
    conductivity : float
        Beam pipe conductivity in :math:`s / m`
    Z0 : float
        Characteristic impedance of vacuum in* :math:`\Omega`

    Examples
    ----------
    >>> pipe_radius = 1
    >>> pipe_length = 2
    >>> resistivity = 3
    >>> rw = ResistiveWall(pipe_radius, pipe_length, resistivity)
    >>> frequency = np.array(1,2,3)
    >>> rw.imped_calc(frequency)
    """

    def __init__(self, pipe_radius, pipe_length, resistivity=None,
                 conductivity=None):

        _ImpedanceObject.__init__(self)

        # Beam pipe radius in m
        self.pipe_radius = float(pipe_radius)

        # Beam pipe length in m
        self.pipe_length = float(pipe_length)

        # Beam pipe conductivity in :math:`s / m`
        if resistivity is not None:
            self.conductivity = 1 / resistivity
        elif conductivity is not None:
            self.conductivity = conductivity
        else:
            # MissingParameterError
            raise RuntimeError('At least one of the following parameters ' +
                               'should be provided: resistivity or conductivity')

        # Characteristic impedance of vacuum in* :math:`\Omega`
        self.Z0 = physical_constants['characteristic impedance of vacuum'][0]

    @property
    def resistivity(self):
        return self.__resistivity

    @resistivity.setter
    def resistivity(self, resistivity):
        self.__resistivity = resistivity
        self.__conductivity = 1 / resistivity

    @property
    def conductivity(self):
        return self.__conductivity

    @conductivity.setter
    def conductivity(self, conductivity):
        self.__resistivity = 1 / conductivity
        self.__conductivity = conductivity

    def imped_calc(self, frequency_array):
        r"""
        Impedance calculation method as a function of frequency.

        Parameters
        ----------
        frequency_array : float array
            Input frequency array in Hz

        Attributes
        ----------
        frequency_array : float array
            Input frequency array in Hz
        impedance : complex array
            Output impedance in :math:`\Omega + j \Omega`
        """

        self.frequency_array = frequency_array

        self.impedance = (self.Z0 * c * self.pipe_length /
                          (np.pi * (1.0 - 1j * np.sign(self.frequency_array)) * 2 *
                           self.pipe_radius * c * np.sqrt(self.conductivity * self.Z0 * c /
                                                          (4.0 * np.pi * np.abs(self.frequency_array)))
                           + 1j * self.pipe_radius**2.0 * 2.0 * np.pi * self.frequency_array)).astype(dtype=bm.precision.complex_t, order='C', copy=False)

        self.impedance[np.isnan(self.impedance)] = 0.0


class CoherentSynchrotronRadiation(_ImpedanceObject):
    r"""
    Impedance contribution of coherent synchrotron radiation emitted insisde a dipole magnet with
    bending raidus :math:`R`.

    The free-space impedance increases as :math:`f^{1/3}` until the critical frequency

    .. math::
        f_{\text{crit}} = \frac{3}{4\pi} \gamma^3 \frac{c}{R},

    and is exponentially surpressed above. Notice that the critical frequency strongly depends on 
    energy, but the low-frequency regime is energy-independent.

    The parallel-plates impedance models the vacuum chamber as a pair of infinite, perfectly 
    conducting plates, separated by a distance `chamber_height` (i.e. distance from top to bottom).
    Frequencies bewlow the cut-off frequency

    .. math::
        f_{\text{cut}} = \sqrt{\frac{2}{3}} \left(\pi \frac{R}{h}\right)^{3/2} \frac{c}{R}\,,

    are exponentially surpressed. For larger frequencies, it approaches the free-space impedance.

    Depending on the input parameters, different models are chosen:

    * Without any optional parameters, the low-frequency approximation to the free-space impedance
      is used, given by eq. 6.18 of [Murphy1997]_
    * If the Lorentz `gamma` is passed as the *only* optional parameter, the exact free-space
      impedance is used, implemented as eqs. A4 and A5 of [Murphy1997]_.
    * If the height of the vacuum chamber (from top to bottom) is passed as the *only* optional
      `chamber_height`, the low-frequency parallel-plates model as given by eq. 8 of
      [Chao2011]_ is used.
    * Finally, if both `gamma` *and* the `chamber_height` are passed, the parallel-plate impedance
      model from eq. B13 in [Murphy1997]_ is used.

    .. image:: csr_impedance_real.png
        :align: right

    The evaluation of the exact impedances is numerically unstable (and time-consuming) for
    frequencies larger than the critical frequency. Therefore,
    the kwarg `high_frequency_transition` sets the frequency (in units of the critical
    frequency) above which a simpler approximate expression is used.

    Note
    ----------
    The (incoherent) energy loss due to synchrotron radiation is *not* included. To include it,
    use :class:`~blond.synchrotron_radiation.synchrotron_radiation.SynchrotronRadiation`.

    References
    ----------
    .. [Murphy1997] J.B. Murphy, S. Krinsky, R.L. Gluckstern, "Longitudinal wakefield for
        an electron moving on a circular orbit", *Particle Accelerators*, vol. 57, p. 9--64, 1997.

    .. [Chao2011] Y. Cai, "Theory of microwave instability and coherent synchrotron radiation
        in electron storage rings", in *Proc. IPAC'11*, San Sebastian, Spain, May 2011,
        pp. 3774--3778

    Examples
    ----------
    >>> from blond.impedances.impedance_sources import CoherentSynchrotronRadiation
    >>> freqs = 10**np.linspace(8,15, num=200)  # frequencies at which to compute impedance

    Specify bending radius and chamber height (both in [m]) and the Lorentz gamma
    for 40MeV electrons

    >>> r_bend, h, gamma = 1.273, 32e-3, 40e6/511e3

    With only the bending radius, only the low-freqeuncy free-space impedance is computed:

    >>> Z_fs_appr = CoherentSynchrotronRadiation(r_bend)
    >>> Z_fs_appr.imped_calc(freqs)

    To compute the free-space impedance, also pass only the Lorentz gamma:

    >>> Z_fs = CoherentSynchrotronRadiation(rBend, gamma=gamma)
    >>> Z_fs.imped_calc(freqs, low_frequency_transition=1e-4, high_frequency_transition=20)

    To compute the approximate parallel-plates impedance, pass the chamber height:

    >>> Z_pp_appr = CoherentSynchrotronRadiation(rBend, chamber_height=h)
    >>> Z_pp_appr.imped_calc(freqs, high_frequency_transition=10)

    For the computation of the parallel-plates impedance you need to pass both the chamber height
    and the Lorentz gamma:

    >>> Z_pp = CoherentSynchrotronRadiation(rBend, chamber_height=h, gamma=gamma)
    >>> Z_pp.imped_calc(freqs, low_frequency_transition=1e-4, high_frequency_transition=20)

    """

    def __init__(self, r_bend, gamma=None, chamber_height=np.inf):
        r"""


        Parameters
        ----------
        r_bend : float
            Bending radius in m
        gamma : float, optional
            The Lorentz factor is only used when the impedance is computed The default is None.
        chamber_height : float, optional
            The height of the vacuum chamber, i.e. the separation between the parallel plates.
            `chamber_height` is required for the
            The default is None.
        low_frequency_approximation : bool, optional
            If true, the low-frequency approximation for either the free-space or parallel
            plates impedances are used. If false, the Lorentz factor `gamma` must be specified.
            The default is True.
        parallel_plates : TYPE, optional
            If ture, the parallel plates impedance is computed. In this case, `chamber_height`
            must be specified. If false, the free-space impedance is computed. The default is False.

        Raises
        ------
        ValueError
            Raised when input parameters are inconsistent, e.g. using the parallel plates impedance,
            but not specifying the `chamber_height`.
        RuntimeError
            Raised when the method for computing the impedance could not be deterimined.

        Returns
        -------
        None.

        """
        _ImpedanceObject.__init__(self)

        # Characteristic impedance of vacuum in* :math:`\Omega`
        self.Z0 = physical_constants['characteristic impedance of vacuum'][0]

        self.r_bend = r_bend
        self.gamma = gamma
        self.chamber_height = chamber_height

        # test for input consistency
        if self.r_bend <= 0.0:
            raise ValueError('bending radius must be greater 0')
        if self.chamber_height <= 0.0:
            raise ValueError('chamber_height must be greater 0')
        if self.gamma is not None and self.gamma <= 1.0:
            raise ValueError('gamma must be greater 1')

        # TODO use f_0 = f_rev (from ring length) or f_0 = c/r_bend/2pi (only path in dipoles)?
        self.f_0 = 0.5 * c / self.r_bend / np.pi  # assumes beta == 1

        if self.chamber_height < np.inf:
            self.Delta = self.chamber_height / self.r_bend
            # parallel plates cut-off frequency
            self.f_cut = np.sqrt(2 / 3) * (np.pi / self.Delta)**1.5 * self.f_0

        if self.gamma is not None:
            # critical frequency in Hz
            self.f_crit = 0.75 * self.gamma**3 * c / self.r_bend / np.pi

            # used for the computation of the free-space impedance
            self.hyper_vec = np.vectorize(mpmath.hyper, excluded=[0, 1], otypes=[float])

        # chose proper impedance method based on input
        if self.gamma is None and np.isinf(self.chamber_height):
            self.imped_calc = self._fs_low_frequency_wrapper

        elif self.gamma is not None and np.isinf(self.chamber_height):
            self.imped_calc = self._fs_spectrum

        elif self.gamma is None and not np.isinf(self.chamber_height):
            self.imped_calc = self._pp_low_frequency

        elif self.gamma is not None and not np.isinf(self.chamber_height):
            self.imped_calc = self._pp_spectrum

        else:
            # WrongCalcError
            raise RuntimeError(
                'method for impedance calculation in CoherentSynchrotronRadiation object '
                + 'not recognized')

    def _pp_low_frequency(self, frequency_array, u_max=10, high_frequency_transition=np.inf):
        """
        Computes the parallel-plates impedance according to eq. 8 of [Chao2011]_. For frequencies
        larger than the cut-off frequencies, it approaches the low-frequency free-space impedance,
        i.e. the impedance increases as :math:`f^{1/3}`.

        Parameters
        ----------
        frequency_array : float, array
            Frequencies at which to compute the impedance.
        u_max : float, optional
            Maximum value of u, for which the summands are evaluated. The Airy functions become
            numerically unstable for u>100. The real part is exponentially supressed, but
            imaginary part is not and the remaining sum from u_max to infinity is
            estimated analytically. The default is 10.
        high_frequency_transition : float, optional
            Use the low-frequency free-space impedance for frequencies above
            `high_frequency_transition` times cut_off frequency. The default is np.inf.

        Raises
        ------
        ValueError
            Raised if `high_frequency_transition` < 1.

        Yields
        ------
        None.

        """

        # using high_frequency_transition factor of 1 yields unphysical results
        if high_frequency_transition <= 1.0:
            raise ValueError('high frequency transition must be greater than 1')

        n_array = frequency_array / self.f_0

        self.impedance = np.zeros_like(n_array, dtype=complex)

        # parallel plates cut-off frequency in units of f_0
        n_cut = np.sqrt(2 / 3) * (np.pi / self.Delta)**1.5

        # use approximate equation for frequencies above high_frequency_transition * n_cut
        approx_indexes = n_array > (high_frequency_transition * n_cut)
        if np.count_nonzero(approx_indexes) > 0:
            self.impedance[approx_indexes] = self._fs_low_frequency(frequency_array[approx_indexes])

        # use exact result for all other frequencies (and ignore f=0)
        exact_indexes = np.invert(approx_indexes) * n_array != 0

        # if there is nothing to calculate, we are done ...
        if np.count_nonzero(exact_indexes) == 0:
            return

        n_array = n_array[exact_indexes]  # override for convenience and exclude f=0

        # maximum p(n) to have u(p,n)<u_max(n)
        pMax_array = np.array(np.ceil(self.Delta * n_array**(2 / 3) * np.sqrt(u_max) / (2**(2 / 3) * np.pi)
                                      - 0.5), dtype=int)

        pMax = pMax_array[-1]  # maximum p; assumes largest frequency is at last array element

        p_matrix = np.zeros(shape=(len(n_array), pMax), dtype=int)

        # matrix to store the summands
        Z_matrix = np.zeros_like(p_matrix, dtype=complex)

        for nit, n in enumerate(n_array):
            # first element of p_matrix is 1 to ensure evaluation at u_min...
            # ... if n is large enough so that u_min < 100, (i.e. airy(u_min) does not yield np.nan)
            if pMax_array[nit] == 0 and n > (np.pi / self.Delta)**1.5 / np.sqrt(2) / 100**0.75:
                p_matrix[nit, 0] = 1
            else:
                p_matrix[nit, :pMax_array[nit]] = (2 * np.arange(pMax_array[nit]) + 1)**2

        # evaluate Airy functions only at these values of p
        indexes = p_matrix > 0

        # argument of the Airy functions
        u_matrix = ((np.pi / 2**(1 / 3) / self.Delta)**2 / n_array**(4 / 3) * p_matrix.T).T

        # evaluate Airy function only at relevant indexes
        airy_matrix = airy(u_matrix[indexes])  # returns Ai(u), Ai'(u), Bi(u), Bi'(u)

        Ci_matrix = airy_matrix[0] - 1j * airy_matrix[2]
        Ci_prime_matrix = airy_matrix[1] - 1j * airy_matrix[3]

        Z_matrix[indexes] = airy_matrix[1] * Ci_prime_matrix \
            + u_matrix[indexes] * airy_matrix[0] * Ci_matrix

        # sum over p
        self.impedance[exact_indexes] = np.sum(Z_matrix, axis=1)

        # the sum in eq. 8 is up to infinity, but we stop at p_max; the real part is exponentially
        # surpressed for large u, but the imaginary part is not; using the assymptotic expression
        # of the imaginary summands for large u, the remaining sum from p_max to infinity can be
        # performed analytically by Mathematica 12.1.0.0.
        self.impedance[exact_indexes] += 1j * 2**(5 / 3) / (4096 * np.pi**6) \
            * (self.Delta * n_array**(2 / 3))**5 * polygamma(4, pMax_array + 1)

        self.impedance[exact_indexes] *= self.Z0 * 4 * np.pi**2 * 2**(1 / 3) \
            * 1 / self.Delta / n_array**(1 / 3)

    def _pp_spectrum(self, frequency_array, zeta_max=9, **kwargs):
        r'''
        Computes the parallel-plates impedance, based on eq. B13 of [Murphy1997]_.

        .. math::
            Z_{pp}(f) = Z_{fs,exact} - Z_{fs,approx} + 
            Z_0 \sqrt{8\pi} 3^{2/3} \exp(i \pi/6) \left(\frac{f}{f_0}\right)^{1/3} 
            \frac{1}{\alpha(f)}
            \sum_{p=0}^\infty h\left(\frac{1}{2}\left(\frac{\pi(2p+1)}{\alpha(f)}\right)^2 \right)

        Parameters
        ----------
        frequency_array : float array
            Frequency at which to calculate parallel-plates impedance
        zeta_max : float, optional
            Maximum value of zeta, for which :meth:`~_hFun` is evaluated. The result becomes
            numerically unstable for zeta>18. The remaining sum from zeta_max to infinity is
            estimated analytically. The default is 9.
        **kwargs :
            Keyword arguments get passed to :meth:`~_fs_spectrum`

        Returns
        -------
        None.

        '''

        non_zero_indexes = frequency_array != 0

        n_array = frequency_array[non_zero_indexes] / self.f_0

        # Murphy et al. convention has plates at +/-h, wheras we use h for the full chamber height
        # (from top to bottom) => we need to use 0.5*self.Delta in their equations

        # based an eq. B8
        alphas = np.sqrt(2) * np.exp(-1j * np.pi / 6) * n_array**(2 / 3) * 0.5 * self.Delta / 3**(1 / 6)

        # sets self.impedance to the full free-space impedance
        self._fs_spectrum(frequency_array, **kwargs)

        # subtract low-frequency free-space impedance
        self.impedance -= self._fs_low_frequency(frequency_array)

        # parallel plates part

        # maximum summation index p(n), such that zeta(p,n) < zeta_max
        pMax_array = np.array(np.ceil(np.sqrt(zeta_max / 3**(1 / 3)) / np.pi * 0.5 * self.Delta
                                      * n_array**(2 / 3) - 0.5), dtype=int)

        pMax = pMax_array[-1]  # maximum p; assumes largest frequency is at last array element

        p_matrix = np.zeros(shape=(len(n_array), pMax), dtype=int)

        # matrix to store the summands
        Z_matrix = np.zeros_like(p_matrix, dtype=complex)

        for nit, n in enumerate(n_array):
            # first element of p_matrix is 1 to ensure evaluation at zeta_min...
            # ... if n is large enough so that zeta_min < zeta_max
            if pMax_array[nit] == 0 and n > 3**0.25 * (np.pi / (0.5 * self.Delta))**1.5 / zeta_max**0.75:
                p_matrix[nit, 0] = 1
            else:
                p_matrix[nit, :pMax_array[nit]] = (2 * np.arange(pMax_array[nit]) + 1)**2

        # evaluate h function only at these values of p
        indexes = p_matrix > 0

        z_matrix = (0.5 * (np.pi / alphas)**2 * p_matrix.T).T

        Z_matrix[indexes] = self._hFun(z_matrix[indexes])

        # sum over p
        Z_pp = np.sum(Z_matrix, axis=1)

        # we cut the infinite sum at p_max; using the assymptotic expression of _hFun for large
        # zeta, the remaining sum from p_max to infinity can be performed analytically by
        # Mathematica 12.1.0.0
        Z_pp += np.sqrt(np.pi) / (32 * 3**(5 / 6)) * np.exp(1j * np.pi / 6)\
            * (0.5 * self.Delta * n_array**(2 / 3) / np.pi)**5\
            * polygamma(4, pMax_array + 0.5)

        Z_pp *= self.Z0 * (8 * np.pi)**0.5 * 3**(2 / 3) * np.exp(1j * np.pi / 6) * n_array**(1 / 3) / alphas

        self.impedance[non_zero_indexes] += Z_pp

    def _hFun(self, z):
        r"""
        Implements eq. B14 of [Murphy1997]_. 

        The integral in eq. B14 was solved analytically by Mathematica 12.1.0.0. However,
        the analytic solution in terms of Airy functions ist numerically unstable for
        :math:`\zeta>18`, with :math:`z=\exp(i\pi/3)\zeta`. Numerically solving eq. B14,
        gives a negligable contribution for :math:`zeta > 18.0`, and the analytic solution is,
        thus, sufficient for our purposes.

        Parameters
        ----------
        z : complex array
            Argument of the function. Numerically unstable for Re z > 9, but this
            condition is not checked.

        Returns
        -------
        complex array
            h(z)

        """

        airy_array = airy(-z / 12**(1 / 3))  # returns Ai(), Ai'(), Bi(), Bi'()
        return - np.pi**1.5 / (2**(2 / 3) * 3**(5 / 6))\
            * (z * (airy_array[0]**2 + airy_array[2]**2) / 12**(1 / 3)
               - airy_array[1]**2 - airy_array[3]**2)

    def _fs_spectrum(self, frequency_array, epsilon=1e-6,
                     low_frequency_transition=1e-5, high_frequency_transition=10):
        r"""
        Computes the exact free-space synchrotron radiation impedance, based on eqs. A4 and A5 of
        [Murphy1997]_. For computation speed and numerical stability, the approximate expressions
        are used for frequencies lower / higher than the critical frequency `f_crit`.

        Parameters
        ----------
        frequency_array : float array
            Frequencies at which to evaluate the impedance
        high_frequency_transition : float, optional
            Ratio of f/f_crit above which the high-frequency approximation is used. If it is smaller
            than 1, a `ValueError` is raised. The default is 10.
        low_frequency_transition : float, optional
            Ratio of f/f_crit below which the low-frequency approximation is used. If it is greater
            than 1, a `ValueError` is raised. The default is 0, i.e. the approximation is not used.
        epsilon : float, optional
            The first integral of eq. A5 has an integrable singularity at the upper limit 1,
            which is (currently) not handled by :func:`~scipy.integrate.quad_vec`.
            Therefore, the integration is only up to 1-`epsilon`.

        Raises
        ------
        ValueError
            Raised if
            * `high_frequency_transition` < 1
            * `low_frequency_transition` > 1
            * `high_frequency_transition` < `low_frequency_transition`

        Returns
        -------
        None.

        """

        # check for input consistency
        if high_frequency_transition <= low_frequency_transition:
            raise ValueError('high_frequency_transition ratio must be larger than the '
                             + 'low_frequency_transition rato')

        if high_frequency_transition < 1.0:
            raise ValueError('high_frequency_transition ratio must be greater than 1')

        if low_frequency_transition > 1.0:
            raise ValueError('low_frequency_transition ratio must be smaller than 1')

        if epsilon < 0 or epsilon > 1:
            raise ValueError('epsilon must be a small positive value')

        self.impedance = np.zeros_like(frequency_array, dtype=complex)

        l_array = frequency_array / self.f_crit

        # use the low frequency approximation where f < LFT * f_c
        # low_indexes = np.where(l_array < low_frequency_transition)[0]
        low_indexes = l_array < low_frequency_transition
        if np.count_nonzero(low_indexes) > 0:
            self.impedance[low_indexes] = self._fs_low_frequency(frequency_array[low_indexes])

        # use the high frequency approximation where f > HFT * f_c
        # high_indexes = np.where(l_array > high_frequency_transition)[0]
        high_indexes = l_array > high_frequency_transition
        if np.count_nonzero(high_indexes) > 0:
            self.impedance[high_indexes] = self._fs_high_frequency(frequency_array[high_indexes])

        # use full integration for frequencies inbetween
        exact_indexes = np.invert(low_indexes + high_indexes)
        if np.count_nonzero(exact_indexes) == 0:
            return

        # Real part: eq. A4, is solved analytically with Mathematica 12.1.0.0 in terms of
        # generalized hypergeometric functions
        # Imaginary part: quad_vec can't handle the integrable singularity at y=1 for y<1, we need
        # to integrate up to 1-epsilon
        self.impedance[exact_indexes] =\
            np.sqrt(3) * gamma_func(2 / 3) / l_array[exact_indexes]**(2 / 3) / 2**(4 / 3)\
            * self.hyper_vec([-1 / 3], [-2 / 3, 2 / 3], 0.25 * l_array[exact_indexes]**2) \
            + 81 * np.pi * l_array[exact_indexes]**(8 / 3) / (640 * 2**(2 / 3) * gamma_func(-1 / 3))\
            * self.hyper_vec([4 / 3], [7 / 3, 8 / 3], 0.25 * l_array[exact_indexes]**2)\
            - 0.25 * np.pi\
            + 1j * (integrate.quad_vec(lambda y: self._fs_integrandImZ1(y, l_array[exact_indexes]),
                                       0, 1 - epsilon)[0]
                    - integrate.quad_vec(lambda y: self._fs_integrandImZ2(y, l_array[exact_indexes]),
                                         1, np.inf)[0])

        self.impedance[exact_indexes] *= self.Z0 * self.gamma * l_array[exact_indexes]

    def _fs_low_frequency_wrapper(self, frequency_array):
        r"""
        Wrapper to compute the free-space low-frequency approximation of the synchrotron
        radiation impedance.

        Parameters
        ----------
        frequency_array : float array
            Frequencies at which to compute the impedance

        Returns
        -------
        None.

        """

        self.impedance = self._fs_low_frequency(frequency_array)

    def _fs_low_frequency(self, frequency_array):
        r"""
        Computes the free-space low-frequency approximation of the synchrotron radiation impedance,
        according to eq. 6.18 of [Murphy1997]_.

        .. math::
          Z_{fs,approx}(f) / Z_0 = \frac{\Gamma\left( 2/3 \right) }{3^{1/3}} e^{i\pi/6}
          \left(\frac{f}{f_0}\right)^{1/3}

        This is a helper function.

        Parameters
        ----------
        frequency_array : float array
            Frequencies at which to compute the impedance

        Returns
        -------
        complex array
            impedance
        """

        return self.Z0 * gamma_func(2 / 3) / 3**(1 / 3) * np.exp(1j * np.pi / 6) \
            * (frequency_array / self.f_0)**(1 / 3)

    def _fs_high_frequency(self, frequency_array):
        r"""
        Computes the free-space high-frequency approximation of the synchrotron radiation impedance,
        based on eq. 6.20 of [Murphy1997]_. This function is a helper function for
        _fs_full_spectrum.

        Parameters
        ----------
        frequency_array : float array
            Frequencies at which to compute the impedance

        Returns
        -------
        complex array
            impedance
        """

        return self.Z0 * self.gamma * (np.sqrt(3 * np.pi / 32)
                                       * np.sqrt(frequency_array / self.f_crit) * bm.exp(-frequency_array / self.f_crit)
                                       - 4j / 9 * self.f_crit / frequency_array)

    def _fs_integrandReZ(self, x):
        # integrand of real part of free-space impedance
        return kv(5 / 3, x)

    def _fs_integrandImZ1(self, y, x):
        # integrand of imaginary part of free-space impedance for y<1
        return np.real(bm.exp(-x * y) * (
            (1j * y + np.sqrt(1 - y**2))**(5 / 3) + 1 / (1j * y + np.sqrt(1 - y**2))**(5 / 3) - 2 * np.sqrt(1 - y**2))
            / (4 * y * np.sqrt(1 - y**2)))

    def _fs_integrandImZ2(self, y, x):
        # integrand of imaginary part of free-space impedance for y>1
        return bm.exp(-x * y) * (
            8 * np.sqrt(y**2 - 1) - (y + np.sqrt(y**2 - 1))**(5 / 3) + 1 / (y + np.sqrt(y**2 - 1))**(5 / 3))\
            / (8 * y * np.sqrt(y**2 - 1))



# coding: utf8
# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''

**Utilities to calculate Hamiltonian, separatrix, total voltage for the full ring.**

:Authors: **Danilo Quartullo**, **Helga Timko**, **Alexandre Lasheen**
'''


from __future__ import division, print_function

import copy
import warnings
from builtins import range

import numpy as np
from scipy.constants import c
import scipy
from packaging.version import Version

if Version(scipy.__version__) >= Version("1.14"):
    from scipy.integrate import cumulative_trapezoid as cumtrapz
else:
    from scipy.integrate import cumtrapz


from ..utils import bmath as bm


def synchrotron_frequency_distribution(Beam, FullRingAndRF, main_harmonic_option='lowest_freq',
                                       turn=0, TotalInducedVoltage=None, smoothOption=None):
    '''
    *Function to compute the frequency distribution of a distribution for a certain
    RF system and optional intensity effects. The potential well (and induced
    potential) are not updated by this function, thus it has to be called* 
    **after** *the potential well (and induced potential) generation.*

    *If used with induced potential, be careful that noise can be an issue. An
    analytical line density can be inputed by using the TotalInducedVoltage 
    option and passing the following parameters:*

    TotalInducedVoltage = beam_generation_output[1]

    *with beam_generation_output being the output of the 
    matched_from_line_density and matched_from_distribution_density functions 
    in the distribution module.*

    *A smoothing function is included (running mean) in order to smooth
    noise and numerical errors due to linear interpolation, the user can input the 
    number of pixels to smooth with smoothOption = N.*

    *The particle distribution in synchrotron frequencies of the beam is also
    outputed.*
    '''

    # Initialize variables depending on the accelerator parameters
    slippage_factor = FullRingAndRF.RingAndRFSection_list[0].rf_params.eta_0[0]

    eom_factor_dE = abs(slippage_factor) / (2 * Beam.beta**2. * Beam.energy)
    eom_factor_potential = np.sign(slippage_factor) * Beam.Particle.charge / (FullRingAndRF.RingAndRFSection_list[0].rf_params.t_rev[0])

    # Generate potential well
    n_points_potential = int(1e4)
    FullRingAndRF.potential_well_generation(n_points=n_points_potential,
                                            turn=turn, dt_margin_percent=0.05,
                                            main_harmonic_option=main_harmonic_option)
    potential_well_array = FullRingAndRF.potential_well
    time_coord_array = FullRingAndRF.potential_well_coordinates

    induced_potential_final = 0

    # Calculating the induced potential
    if TotalInducedVoltage is not None:

        induced_voltage_object = copy.deepcopy(TotalInducedVoltage)

        induced_voltage = induced_voltage_object.induced_voltage
        time_induced_voltage = TotalInducedVoltage.profile.bin_centers

        # Computing induced potential
        induced_potential = - eom_factor_potential * np.insert(cumtrapz(induced_voltage, dx=time_induced_voltage[1] - time_induced_voltage[0]), 0, 0)

        # Interpolating the potential well
        induced_potential_final = np.interp(time_coord_array, time_induced_voltage, induced_potential)

    # Induced voltage contribution
    total_potential = potential_well_array + induced_potential_final

    # Process the potential well in order to take a frame around the separatrix
    time_coord_sep, potential_well_sep = potential_well_cut(time_coord_array, total_potential)

    potential_well_sep = potential_well_sep - np.min(potential_well_sep)
    synchronous_phase_index = np.where(potential_well_sep == np.min(potential_well_sep))[0]

    # Computing the action J by integrating the dE trajectories
    J_array_dE0 = np.zeros(len(potential_well_sep))

    warnings.filterwarnings("ignore")

    for i in range(0, len(potential_well_sep)):
        # Find left and right time coordinates for a given hamiltonian
        # value
        time_indexes = np.where(potential_well_sep <=
                                potential_well_sep[i])[0]
        left_time = time_coord_sep[np.max((0, time_indexes[0]))]
        right_time = time_coord_sep[np.min((time_indexes[-1],
                                            len(time_coord_sep) - 1))]
        # Potential well calculation with high resolution in that frame
        time_potential_high_res = np.linspace(float(left_time), float(right_time),
                                              n_points_potential)
        FullRingAndRF.potential_well_generation(
            n_points=n_points_potential,
            time_array=time_potential_high_res,
            main_harmonic_option=main_harmonic_option)
        pot_well_high_res = FullRingAndRF.potential_well
        if TotalInducedVoltage is not None:
            pot_well_high_res += np.interp(time_potential_high_res,
                                           time_induced_voltage, induced_potential)
            pot_well_high_res -= pot_well_high_res.min()
        # Integration to calculate action
        dE_trajectory = np.sqrt((potential_well_sep[i] -
                                 pot_well_high_res) / eom_factor_dE)
        dE_trajectory[np.isnan(dE_trajectory)] = 0
        J_array_dE0[i] = 1 / np.pi * np.trapz(dE_trajectory,
                                              dx=time_potential_high_res[1] - time_potential_high_res[0])

    warnings.filterwarnings("default")

    # Computing the sync_freq_distribution (if to handle cases where maximum is in 2 consecutive points)
    if len(synchronous_phase_index) > 1:
        H_array_left = potential_well_sep[0:synchronous_phase_index[0] + 1]
        H_array_right = potential_well_sep[synchronous_phase_index[1]:]
        J_array_left = J_array_dE0[0:synchronous_phase_index[0] + 1]
        J_array_right = J_array_dE0[synchronous_phase_index[1]:]
        delta_time_left = time_coord_sep[0:synchronous_phase_index[0] + 1]
        delta_time_right = time_coord_sep[synchronous_phase_index[1]:]
        synchronous_time = np.mean(time_coord_sep[synchronous_phase_index])
    else:
        H_array_left = potential_well_sep[0:synchronous_phase_index[0] + 1]
        H_array_right = potential_well_sep[synchronous_phase_index[0]:]
        J_array_left = J_array_dE0[0:synchronous_phase_index[0] + 1]
        J_array_right = J_array_dE0[synchronous_phase_index[0]:]
        delta_time_left = time_coord_sep[0:synchronous_phase_index[0] + 1]
        delta_time_right = time_coord_sep[synchronous_phase_index[0]:]
        synchronous_time = time_coord_sep[synchronous_phase_index]

    delta_time_left = delta_time_left[-1] - delta_time_left
    delta_time_right = delta_time_right - delta_time_right[0]

    if smoothOption is not None:
        H_array_left = np.convolve(H_array_left, np.ones(smoothOption) / smoothOption, mode='valid')
        J_array_left = np.convolve(J_array_left, np.ones(smoothOption) / smoothOption, mode='valid')
        H_array_right = np.convolve(H_array_right, np.ones(smoothOption) / smoothOption, mode='valid')
        J_array_right = np.convolve(J_array_right, np.ones(smoothOption) / smoothOption, mode='valid')
        delta_time_left = (delta_time_left + (smoothOption - 1) * (delta_time_left[1] - delta_time_left[0]) / 2)[0:len(delta_time_left) - smoothOption + 1]
        delta_time_right = (delta_time_right + (smoothOption - 1) * (delta_time_right[1] - delta_time_right[0]) / 2)[0:len(delta_time_right) - smoothOption + 1]

    delta_time_left = np.fliplr([delta_time_left])[0]

    # Calculation of fs as fs= dH/dJ / (2*pi)
    sync_freq_distribution_left = np.gradient(H_array_left) / np.gradient(J_array_left) / (2 * np.pi)
    sync_freq_distribution_left = np.fliplr([sync_freq_distribution_left])[0]
    sync_freq_distribution_right = np.gradient(H_array_right) / np.gradient(J_array_right) / (2 * np.pi)

    # Emittance arrays
    emittance_array_left = J_array_left * (2 * np.pi)
    emittance_array_left = np.fliplr([emittance_array_left])[0]
    emittance_array_right = J_array_right * (2 * np.pi)

    # Calculating particle distribution in synchrotron frequency
    H_particles = eom_factor_dE * Beam.dE**2 + np.interp(Beam.dt, time_coord_array, total_potential)
    sync_freq_distribution = np.concatenate((sync_freq_distribution_left, sync_freq_distribution_right))
    H_array = np.concatenate((np.fliplr([H_array_left])[0], H_array_right))
    sync_freq_distribution = sync_freq_distribution[H_array.argsort()]
    H_array.sort()

    particleDistributionFreq = np.interp(H_particles, H_array, sync_freq_distribution)

    return [sync_freq_distribution_left, sync_freq_distribution_right], \
        [emittance_array_left, emittance_array_right], \
        [delta_time_left, delta_time_right], \
        particleDistributionFreq, synchronous_time


class synchrotron_frequency_tracker:
    '''
    *This class can be added to the tracking map to track a certain
    number of particles (defined by the user) and to store the evolution
    of their coordinates in phase space in order to compute their synchrotron
    frequency as a function of their amplitude in theta.*

    *As the time step between two turns can change with acceleration, make sure
    that the momentum program is set to be constant when using this function,
    or that beta_rel~1.*

    *The user can input the minimum and maximum theta for the theta_coordinate_range
    option as [min, max]. The input n_macroparticles will be generated with
    linear spacing between these values. One can also input the theta_coordinate_range
    as the coordinates of all particles, but the length of the array should
    match the n_macroparticles value.*
    '''

    def __init__(self, Ring, n_macroparticles, theta_coordinate_range, FullRingAndRF,
                 TotalInducedVoltage=None):

        #: *Number of macroparticles used in the synchrotron_frequency_tracker method*
        self.n_macroparticles = int(n_macroparticles)

        #: *Copy of the input FullRingAndRF object to retrieve the accelerator programs*
        self.FullRingAndRF = copy.deepcopy(FullRingAndRF)

        #: *Copy of the input TotalInducedVoltage object to retrieve the intensity effects
        #: (the synchrotron_frequency_tracker particles are not contributing to the
        #: induced voltage).*
        self.TotalInducedVoltage = None
        if TotalInducedVoltage is not None:
            self.TotalInducedVoltage = TotalInducedVoltage
            intensity = TotalInducedVoltage.profiles.Beam.intensity
        else:
            intensity = 0.

        #: *Beam object containing the same physical information as the real beam,
        #: but containing only the coordinates of the particles for which the
        #: synchrotron frequency are computed.*
        from ..beam.beam import Beam
        self.Beam = Beam(Ring, n_macroparticles, intensity)

        # Ring object for the ring radius
        self.Ring = Ring

        # Generating the distribution from the user input
        if len(theta_coordinate_range) == 2:
            self.Beam.dt = np.linspace(float(theta_coordinate_range[0]),
                                       float(theta_coordinate_range[1]), n_macroparticles) \
                * (self.Ring.ring_radius / (self.Beam.beta * c))
        else:
            if len(theta_coordinate_range) != n_macroparticles:
                # SynchrotronMotionError
                raise RuntimeError('The input n_macroparticles does not match with the length of the theta_coordinates')
            else:
                self.Beam.dt = np.array(theta_coordinate_range) * (self.Ring.ring_radius / (self.Beam.beta * c))

        self.Beam.dE = np.zeros(int(n_macroparticles))

        for RFsection in self.FullRingAndRF.RingAndRFSection_list:
            RFsection.beam = self.Beam

        #: *Revolution period in [s]*
        self.timeStep = Ring.t_rev[0]

        #: *Number of turns of the simulation (+1 to include the input parameters)*
        self.nTurns = Ring.n_turns + 1

        #: *Saving the theta coordinates of the particles while tracking*
        self.theta_save = np.zeros((self.nTurns, int(n_macroparticles)))

        #: *Saving the dE coordinates of the particles while tracking*
        self.dE_save = np.zeros((self.nTurns, int(n_macroparticles)))

        #: *Tracking counter*
        self.counter = 0

        # The first save coordinates are the input coordinates
        self.theta_save[self.counter] = self.Beam.dt / (self.Ring.ring_radius / (self.Beam.beta * c))
        self.dE_save[self.counter] = self.Beam.dE

    def track(self):
        '''
        *Method to track the particles with or without intensity effects.*
        '''

        self.FullRingAndRF.track()

        if self.TotalInducedVoltage is not None:
            self.TotalInducedVoltage.track_ghosts_particles(self.Beam)

        self.counter = self.counter + 1

        self.theta_save[self.counter] = self.Beam.dt / (self.Ring.ring_radius / (self.Beam.beta * c))
        self.dE_save[self.counter] = self.Beam.dE

    def frequency_calculation(self, n_sampling=100000, start_turn=None, end_turn=None):
        '''
        *Method to compute the fft of the particle oscillations in theta and dE
        to obtain their synchrotron frequencies. The particles for which
        the amplitude of oscillations is extending the minimum and maximum
        theta from user input are considered to be lost and their synchrotron
        frequencies are not calculated.*
        '''

        n_sampling = int(n_sampling)

        #: *Saving the synchrotron frequency from the theta oscillations for each particle*
        self.frequency_theta_save = np.zeros(int(self.n_macroparticles))

        #: *Saving the synchrotron frequency from the dE oscillations for each particle*
        self.frequency_dE_save = np.zeros(int(self.n_macroparticles))

        #: *Saving the maximum of oscillations in theta for each particle
        #: (theta amplitude on the right side of the bunch)*
        self.max_theta_save = np.zeros(int(self.n_macroparticles))

        #: *Saving the minimum of oscillations in theta for each particle
        #: (theta amplitude on the left side of the bunch)*
        self.min_theta_save = np.zeros(int(self.n_macroparticles))

        # Maximum theta for which the particles are considered to be lost
        max_theta_range = np.max(self.theta_save[0, :])

        # Minimum theta for which the particles are considered to be lost
        min_theta_range = np.min(self.theta_save[0, :])

        #: *Frequency array for the synchrotron frequency distribution*
        self.frequency_array = np.fft.rfftfreq(n_sampling, self.timeStep)

        if start_turn is None:
            start_turn = 0

        if end_turn is None:
            end_turn = self.nTurns + 1

        # Computing the synchrotron frequency of each particle from the maximum
        # peak of the FFT.
        for indexParticle in range(0, self.n_macroparticles):
            self.max_theta_save[indexParticle] = np.max(self.theta_save[start_turn:end_turn, indexParticle])
            self.min_theta_save[indexParticle] = np.min(self.theta_save[start_turn:end_turn, indexParticle])

            if (self.max_theta_save[indexParticle] < max_theta_range) and (
                    self.min_theta_save[indexParticle] > min_theta_range):
                theta_save_fft = abs(np.fft.rfft(self.theta_save[start_turn:end_turn, indexParticle] - np.mean(
                    self.theta_save[start_turn:end_turn, indexParticle]), n_sampling))
                dE_save_fft = abs(np.fft.rfft(self.dE_save[start_turn:end_turn, indexParticle] - np.mean(
                    self.dE_save[start_turn:end_turn, indexParticle]), n_sampling))

                self.frequency_theta_save[indexParticle] = self.frequency_array[
                    np.argmax(theta_save_fft == np.max(theta_save_fft))]
                self.frequency_dE_save[indexParticle] = self.frequency_array[
                    np.argmax(dE_save_fft == np.max(dE_save_fft))]


def total_voltage(RFsection_list, harmonic='first'):
    """
    Total voltage from all the RF stations and systems in the ring.
    To be generalized.
    """

    n_sections = len(RFsection_list)

    #: *Sums up only the voltage of the first harmonic RF,
    #: taking into account relative phases*
    if harmonic == 'first':
        Vcos = RFsection_list[0].voltage[0] * np.cos(RFsection_list[0].phi_rf[0])
        Vsin = RFsection_list[0].voltage[0] * np.sin(RFsection_list[0].phi_rf[0])
        if n_sections > 1:
            for i in range(1, n_sections):
                Vcos += RFsection_list[i].voltage[0] * np.cos(RFsection_list[i].phi_rf[0])
                Vsin += RFsection_list[i].voltage[0] * np.sin(RFsection_list[i].phi_rf[0])
        Vtot = np.sqrt(Vcos**2 + Vsin**2)
        return Vtot

    #: *To be implemented*
    elif harmonic == "all":
        return 0

    else:
        warnings.filterwarnings("once")
        warnings.warn("WARNING: In total_voltage, harmonic choice not recognize!")


def hamiltonian(Ring, RFStation, Beam, dt, dE,
                total_voltage=None):
    """Single RF sinusoidal Hamiltonian.
    For the time being, for single RF section only or from total voltage.
    Uses beta, energy averaged over the turn.
    To be generalized."""

    warnings.filterwarnings("once")

    if Ring.n_sections > 1:
        warnings.warn("WARNING: The Hamiltonian is not yet properly computed for several sections!")
    if RFStation.n_rf > 1:
        warnings.warn("WARNING: The Hamiltonian will be calculated for the first harmonic only!")

    counter = RFStation.counter[0]
    h0 = RFStation.harmonic[0, counter]
    if total_voltage is None:
        V0 = float(RFStation.voltage[0, counter])
    else:
        V0 = float(total_voltage[counter])
    V0 *= RFStation.Particle.charge

    c1 = RFStation.eta_tracking(Beam, counter, dE) * c * np.pi / \
        (Ring.ring_circumference * Beam.beta * Beam.energy)
    c2 = c * Beam.beta * V0 / (h0 * Ring.ring_circumference)

    phi_s = RFStation.phi_s[counter]
    phi_b = RFStation.omega_rf[0, counter] * dt + \
        RFStation.phi_rf_d[0, counter]

    eta0 = RFStation.eta_0[counter]

    # Modulo 2 Pi of bunch phase
    if eta0 < 0:
        phi_b = phase_modulo_below_transition(phi_b)
    elif eta0 > 0:
        phi_b = phase_modulo_above_transition(phi_b)

    return c1 * dE**2 + c2 * (bm.cos(phi_b) - bm.cos(phi_s) +
                              (phi_b - phi_s) * bm.sin(phi_s))


def separatrix(Ring, RFStation, dt):
#TODO:  Use list of RFStation and consider all voltages instead of multiplying by n sections
    r""" Function to calculate the ideal separatrix without intensity effects.
    For single or multiple RF systems. For the time being, multiple RF sections
    are implemented for the case that all RF stations have the same voltage over
    one turn.

    Parameters
    ---------- 
    Ring : class
        A Ring type class
    RFStation : class
        An RFStation type class
    dt : float array
        Time coordinates the separatrix is to be calculated for

    Returns
    -------
    float array
        Energy coordinates of the separatrix corresponding to dt

    """

    warnings.filterwarnings("once")

    if Ring.n_sections > 1:
        warnings.warn("WARNING in separatrix(): the usage of several RF" +
                      " sections is only implemented for equal energy gains" + 
                      " per RF station per turn!")

    # Import RF and ring parameters at this moment
    counter = RFStation.counter[0]
    voltage = Ring.Particle.charge * RFStation.voltage[:, counter] * Ring.n_sections
    omega_rf = RFStation.omega_rf[:, counter]
    phi_rf = RFStation.phi_rf[:, counter]
    if Ring.Particle.charge < 0:
        phi_rf = phi_rf - np.pi
    eta_0 = RFStation.eta_0[counter]
    beta_sq = RFStation.beta[counter]**2
    energy = RFStation.energy[counter]
    try:
        delta_E = RFStation.delta_E[counter] * Ring.n_sections
    except Exception:
        delta_E = RFStation.delta_E[-1] * Ring.n_sections
    T_0 = Ring.t_rev[counter]
    if Ring.Particle.charge > 0:
        index = np.min(np.where(voltage > 0)[0])
    elif Ring.Particle.charge < 0:
        index = np.max(np.where(voltage < 0)[0])
    T_rf_0 = 2 * np.pi / omega_rf[index]

    # Projects time array into the range [t_RF, t_RF+T_RF] below and above
    # transition, where T_RF = 2*pi/omega_RF, t_RF = - phi_RF/omega_RF.
    # Note that the RF wave is shifted by Pi for eta < 0
    if eta_0 < 0:
        dt = time_modulo(dt, (phi_rf[0] - np.pi) / omega_rf[0],
                         2. * np.pi / omega_rf[0])
    elif eta_0 > 0:
        dt = time_modulo(dt, phi_rf[0] / omega_rf[0], 2. * np.pi / omega_rf[0])

    # Unstable fixed point in single-harmonic RF system
    if RFStation.n_rf == 1:
        if Ring.Particle.charge > 0:
            dt_s = RFStation.phi_s[counter] / omega_rf[0]
        elif Ring.Particle.charge < 0:
            dt_s = (np.pi-RFStation.phi_s[counter]) / omega_rf[0]
            
        if eta_0 < 0:
            dt_RF = -(phi_rf[0] - np.pi) / omega_rf[0]
        else:
            dt_RF = -phi_rf[0] / omega_rf[0]

        dt_ufp = dt_RF + 0.5 * T_rf_0 - dt_s
        if eta_0 * delta_E < 0:
            dt_ufp += T_rf_0

    # Unstable fixed point in multi-harmonic RF system
    else:

        dt_ufp = np.linspace(-float(phi_rf[index] / omega_rf[index] - T_rf_0 / 1000),
                             float(T_rf_0 - phi_rf[index] / omega_rf[index] + T_rf_0 / 1000), 1002)

        if eta_0 < 0:
            dt_ufp += 0.5 * T_rf_0  # Shift in RF phase below transition
        Vtot = np.zeros(len(dt_ufp))

        # Construct waveform
        for i in range(RFStation.n_rf):
            Vtot += voltage[i] * np.sin(omega_rf[i] * dt_ufp + phi_rf[i])
        Vtot -= delta_E

        # Find zero crossings
        zero_crossings = np.where(np.diff(np.sign(Vtot)))[0]

        # Interpolate UFP
        if eta_0 < 0:
            i = -1
            ind = zero_crossings[i]
            while (Vtot[ind + 1] - Vtot[ind]) > 0:
                i -= 1
                ind = zero_crossings[i]
        else:
            i = 0
            ind = zero_crossings[i]
            while (Vtot[ind + 1] - Vtot[ind]) < 0:
                i += 1
                ind = zero_crossings[i]
        dt_ufp = dt_ufp[ind] + Vtot[ind] / (Vtot[ind] - Vtot[ind + 1]) * \
            (dt_ufp[ind + 1] - dt_ufp[ind])

    # Construct separatrix
    Vtot = np.zeros(len(dt))
    for i in range(RFStation.n_rf):
        Vtot += voltage[i] * (np.cos(omega_rf[i] * dt_ufp + phi_rf[i]) -
                              np.cos(omega_rf[i] * dt + phi_rf[i])) / omega_rf[i]
        
    Vtot *= np.sign(Ring.Particle.charge)
        
    separatrix_sq = 2 * beta_sq * energy / (eta_0 * T_0) * (Vtot + delta_E * (dt_ufp - dt))
    pos_ind = np.where(separatrix_sq >= 0)[0]
    separatrix_array = np.full((len(separatrix_sq)), fill_value=np.nan)
    separatrix_array[pos_ind] = np.sqrt(separatrix_sq[pos_ind])

    return separatrix_array


def is_in_separatrix(Ring, RFStation, Beam, dt, dE,
                     total_voltage=None):
    r"""Function checking whether coordinate pair(s) are inside the separatrix. 
    Uses the single-RF sinusoidal Hamiltonian.

    Parameters
    ---------- 
    Ring : class
        A Ring type class
    RFStation : class
        An RFStation type class
    Beam : class
        A Beam type class
    dt : float array
        Time coordinates of the particles to be checked
    dE : float array
        Energy coordinates of the particles to be checked
    total_voltage : float array
        Total voltage to be used if not single-harmonic RF

    Returns
    -------
    bool array
        True/False array for the given coordinates

    """

    warnings.filterwarnings("once")

    if Ring.n_sections > 1:
        warnings.warn("WARNING: in is_in_separatrix(): the usage of several" +
                      " sections is not yet implemented!")
    if RFStation.n_rf > 1:
        warnings.warn("WARNING in is_in_separatrix(): taking into account" +
                      " the first harmonic only!")

    counter = RFStation.counter[0]
    dt_sep = (np.pi - RFStation.phi_s[counter]
              - RFStation.phi_rf_d[0, counter]) / \
        RFStation.omega_rf[0, counter]

    Hsep = hamiltonian(Ring, RFStation, Beam, dt_sep, 0,
                       total_voltage=None)
    isin = bm.fabs(hamiltonian(Ring, RFStation, Beam,
                               dt, dE, total_voltage=None)) < bm.fabs(Hsep)

    return isin


def minmax_location(x, f):
    '''
    *Function to locate the minima and maxima of the f(x) numerical function.*
    '''

    f_derivative = np.diff(f)
    x_derivative = x[0:-1] + (x[1] - x[0]) / 2
    f_derivative = np.interp(x, x_derivative, f_derivative)

    f_derivative_second = np.diff(f_derivative)
    f_derivative_second = np.interp(x, x_derivative, f_derivative_second)

    warnings.filterwarnings("ignore")
    f_derivative_zeros = np.unique(np.append(np.where(f_derivative == 0), np.where(f_derivative[1:] / f_derivative[0:-1] < 0)))

    min_x_position = (x[f_derivative_zeros[f_derivative_second[f_derivative_zeros] > 0] + 1] + x[f_derivative_zeros[f_derivative_second[f_derivative_zeros] > 0]]) / 2
    max_x_position = (x[f_derivative_zeros[f_derivative_second[f_derivative_zeros] < 0] + 1] + x[f_derivative_zeros[f_derivative_second[f_derivative_zeros] < 0]]) / 2

    min_values = np.interp(min_x_position, x, f)
    max_values = np.interp(max_x_position, x, f)

    warnings.filterwarnings("default")

    return [min_x_position, max_x_position], [min_values, max_values]


def potential_well_cut(time_potential, potential_array):
    '''
    *Function to cut the potential well in order to take only the separatrix
    (several cases according to the number of min/max).*
    '''

    # Check for the min/max of the potential well
    minmax_positions, minmax_values = minmax_location(time_potential,
                                                      potential_array)
    min_time_positions = minmax_positions[0]
    max_time_positions = minmax_positions[1]
    max_potential_values = minmax_values[1]
    n_minima = len(min_time_positions)
    n_maxima = len(max_time_positions)

    if n_minima == 0:
        # PotentialWellError
        raise RuntimeError('The potential well has no minima...')
    if n_minima > n_maxima and n_maxima == 1:
        # PotentialWellError
        raise RuntimeError('The potential well has more minima than maxima, and only one maximum')
    if n_maxima == 0:
        print('Warning: The maximum of the potential well could not be found... \
                You may reconsider the options to calculate the potential well \
                as the main harmonic is probably not the expected one. \
                You may also increase the percentage of margin to compute \
                the potentiel well. The full potential well will be taken')
    elif n_maxima == 1:
        if min_time_positions[0] > max_time_positions[0]:
            saved_indexes = (potential_array < max_potential_values[0]) * \
                            (time_potential > max_time_positions[0])
            time_potential_sep = time_potential[saved_indexes]
            potential_well_sep = potential_array[saved_indexes]
            if potential_array[-1] < potential_array[0]:
                # PotentialWellError
                raise RuntimeError('The potential well is not well defined. \
                                    You may reconsider the options to calculate \
                                    the potential well as the main harmonic is \
                                    probably not the expected one.')
        else:
            saved_indexes = (potential_array < max_potential_values[0]) * \
                            (time_potential < max_time_positions[0])
            time_potential_sep = time_potential[saved_indexes]
            potential_well_sep = potential_array[saved_indexes]
            if potential_array[-1] > potential_array[0]:
                # PotentialWellError
                raise RuntimeError('The potential well is not well defined. \
                                    You may reconsider the options to calculate \
                                    the potential well as the main harmonic is \
                                    probably not the expected one.')
    elif n_maxima == 2:
        lower_maximum_value = np.min(max_potential_values)
        higher_maximum_value = np.max(max_potential_values)
        lower_maximum_time = max_time_positions[max_potential_values == lower_maximum_value]
        higher_maximum_time = max_time_positions[max_potential_values == higher_maximum_value]
        if len(lower_maximum_time) == 2:
            saved_indexes = (potential_array < lower_maximum_value) * \
                            (time_potential > lower_maximum_time[0]) * \
                            (time_potential < lower_maximum_time[1])
            time_potential_sep = time_potential[saved_indexes]
            potential_well_sep = potential_array[saved_indexes]
        elif min_time_positions[0] > lower_maximum_time:
            saved_indexes = (potential_array < lower_maximum_value) * \
                            (time_potential > lower_maximum_time) * \
                            (time_potential < higher_maximum_time)
            time_potential_sep = time_potential[saved_indexes]
            potential_well_sep = potential_array[saved_indexes]
        else:
            saved_indexes = (potential_array < lower_maximum_value) * \
                            (time_potential < lower_maximum_time) * \
                            (time_potential > higher_maximum_time)
            time_potential_sep = time_potential[saved_indexes]
            potential_well_sep = potential_array[saved_indexes]
    elif n_maxima > 2:
        left_max_time = np.min(max_time_positions)
        right_max_time = np.max(max_time_positions)
        left_max_value = max_potential_values[max_time_positions == left_max_time]
        right_max_value = max_potential_values[max_time_positions == right_max_time]
        separatrix_value = np.min([left_max_value, right_max_value])
        saved_indexes = (time_potential > left_max_time) * (time_potential < right_max_time) * (potential_array < separatrix_value)
        time_potential_sep = time_potential[saved_indexes]
        potential_well_sep = potential_array[saved_indexes]

    return time_potential_sep, potential_well_sep


def phase_modulo_above_transition(phi):
    '''
    *Projects a phase array into the range -Pi/2 to +3*Pi/2.*
    '''

    return phi - 2. * np.pi * bm.floor(phi / (2. * np.pi))


def phase_modulo_below_transition(phi):
    '''
    *Projects a phase array into the range -Pi/2 to +3*Pi/2.*
    '''

    return phi - 2. * np.pi * (bm.floor(phi / (2. * np.pi) + 0.5))


def time_modulo(dt, dt_offset, T):
    '''
    *Returns dt projected onto the desired interval.*
    '''

    return dt - T * bm.floor((dt + dt_offset) / T)




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Class to choose plots and customize plot layout**

:Authors: **Helga Timko**
'''

import os

import h5py as hp
import matplotlib.pyplot as plt
import numpy as np
from ..plots.plot_beams import (plot_long_phase_space, plot_bunch_length_evol,
                                plot_bunch_length_evol_gaussian, plot_position_evol,
                                plot_energy_evol, plot_transmitted_particles)
from ..plots.plot_llrf import (plot_PL_bunch_phase,
                               plot_PL_RF_phase, plot_PL_phase_corr, plot_PL_RF_freq,
                               plot_PL_freq_corr, plot_RF_phase_error, plot_RL_radial_error,
                               plot_COM_motion, plot_LHCNoiseFB, plot_LHCNoiseFB_FWHM,
                               plot_LHCNoiseFB_FWHM_bbb)
from ..plots.plot_slices import (plot_beam_profile, plot_beam_spectrum)

def fig_folder(dirname):
    '''
    Create folder where plots will be stored.
    '''

    # Try to create directory
    try:
        os.makedirs(dirname)
    # Check whether already exists/creation failed
    except OSError:
        if os.path.exists(dirname):
            pass
        else:
            raise


class Plot:

    def __init__(self, Ring, RFStation, Beam, dt_plot,
                 dt_bckp, xmin, xmax, ymin, ymax, xunit = 's', sampling = 1,
                 show_plots = False,
                 separatrix_plot = False, histograms_plot = True, 
                 Profile = None, h5file = None, output_frequency = 1, 
                 PhaseLoop = None, LHCNoiseFB = None, format_options = None):
        '''
        Define what plots should be plotted during the simulation. Passing only
        basic objects, only phase space plot will be produced. Passing optional
        objects, plots related to those objects will be produced as well.
        For plots at a certain turn: use 'dt_plot' to set the plotting frequency 
        in units of time steps. 
        For plots as a function of time: use 'dt_bckp' to set plotting frequency
        in units of time steps.
        '''


        #: | *Import Ring*
        self.general_params = Ring

        #: | *Import RFStation*
        self.rf_params = RFStation

        #: | *Import actual time step RFStation*
        self.tstep = RFStation.counter

        #: | *Import Beam*
        self.beam = Beam

        #: | *Defining whether the plots should be saved or directly shown*
        self.show_plt = show_plots

        #: | *Plotting frequency in units of time steps*
        self.dt_plot = dt_plot
        self.dt_bckp = dt_bckp

        #: | *Plot limit (where applicable) minimum on x-axis [xunit]*
        self.xmin = xmin

        #: | *Plot limit (where applicable) maximum on x-axis [xunit]*
        self.xmax = xmax

        #: | *Choice of x-axis unit (where applicable) 's' or 'rad'*
        self.xunit = xunit

        #: | *Plot limit (where applicable) minimum on y-axis [eV]*
        self.ymin = ymin

        #: | *Plot limit (where applicable) maximum on y-axis [eV]*
        self.ymax = ymax

        #: | *Sampling of large arrays (where applicable)*
        self.sampling = sampling

        #: | *Separatrix in phase space plot 'True' or 'False'*
        self.separatix = separatrix_plot

        #: | *Histogram in phase space plot 'True' or 'False'*
        self.histogram = histograms_plot

        #: | *Optional import of Profile*
        self.profile = Profile

        #: | *Optional import of Monitor file*
        self.h5file = h5file

        #: | *Optional sampling of monitored quantities*
        self.dt_mon = output_frequency

        #: | *Optional import of PhaseLoop*
        self.PL = PhaseLoop

        #: | *Optional import of LHCNoiseFB*
        self.noiseFB = LHCNoiseFB

        # Set plotting format
        self.set_format(format_options)

        # Track at initialisation
        self.track()

    def set_format(self, format_options):
        '''
        Initialize plot folder and custom plot formatting. For more options, see

        http://matplotlib.org/1.3.1/users/customizing.html
        '''

        if format_options is None:
            format_options = {'dummy': 0}

        if 'dirname' not in format_options:
            self.dirname = 'fig'
        else:
            self.dirname = format_options['dirname']

        if 'linewidth' not in format_options:
            self.lwidth = 2
        else:
            self.lwidth = format_options['linewidth']

        if 'linestyle' not in format_options:
            self.lstyle = '-'
        else:
            self.lstyle = format_options['linestyle']

        if 'markersize' not in format_options:
            self.msize = 6
        else:
            self.msize = format_options['markersize']

        if 'alpha' not in format_options:
            self.alpha = alpha=10**(-np.log10(self.beam.n_macroparticles)/6)
        else:
            self.alpha = format_options['alpha']

        if 'labelsize' not in format_options:
            self.lsize = 18
        else:
            self.lsize = format_options['labelsize']

        if 'fontfamily' not in format_options:
            self.ffamily = 'sans-serif'
        else:
            self.ffamily = format_options['fontfamily']

        if 'fontweight' not in format_options:
            self.fweight = 'normal'
        else:
            self.fweight = format_options['fontweight']

        if 'dpi' not in format_options:
            self.dpi = 100
        else:
            self.dpi = format_options['dpi']

        # Directory where longitudinal_plots will be stored
        fig_folder(self.dirname)

        # Ticksize
        self.tsize = self.lsize - 2

        # Set size of x- and y-grid numbers
        plt.rc('xtick', labelsize=self.tsize)
        plt.rc('ytick', labelsize=self.tsize)

        # Set x- and y-grid labelsize and weight
        plt.rc('axes', labelsize=self.lsize)
        plt.rc('axes', labelweight=self.fweight)

        # Set linewidth for continuous, markersize for discrete plotting
        plt.rc('lines', linewidth=self.lwidth, markersize=self.msize)

        # Set figure resolution, font
        plt.rc('figure', dpi=self.dpi)
        plt.rc('figure', autolayout = True)
        plt.rc('savefig', dpi=self.dpi)
        plt.rc('savefig', bbox='tight')
        plt.rc('savefig', pad_inches = 0.1)
        plt.rc('font', family=self.ffamily)

    def track(self):
        '''
        Plot in certain time steps and depending on imported objects
        '''

        # Snapshot-type plots
        if (self.tstep[0] % self.dt_plot) == 0:
            
            plot_long_phase_space(self.general_params, self.rf_params, 
                                  self.beam, self.xmin, self.xmax, self.ymin, 
                                  self.ymax, self.xunit, 
                                  sampling = self.sampling, 
                                  separatrix_plot = self.separatix, 
                                  histograms_plot = self.histogram, 
                                  dirname = self.dirname, show_plot=self.show_plt,
                                  alpha = self.alpha)
            
            if self.profile:
                plot_beam_profile(self.profile, self.tstep[0], 
                                  style = self.lstyle, dirname = self.dirname, show_plot = self.show_plt)

                self.profile.beam_spectrum_freq_generation(
                    self.profile.n_slices)
                self.profile.beam_spectrum_generation(self.profile.n_slices)
                plot_beam_spectrum(self.profile, self.tstep[0], 
                                   style = self.lstyle, dirname = self.dirname, show_plot = self.show_plt)
        
        # Plots as a function of time        
        if (self.tstep[0] % self.dt_bckp) == 0 and self.h5file:

            h5data = hp.File(self.h5file + '.h5', 'r')
            plot_bunch_length_evol(self.rf_params, h5data, 
                                   output_freq = self.dt_mon, 
                                   dirname = self.dirname, show_plot=self.show_plt)
            
            if self.profile and self.profile.fit_option == 'gaussian':
                    plot_bunch_length_evol_gaussian(self.rf_params, self.profile,
                                                    h5data, 
                                                    output_freq = self.dt_mon, 
                                                    dirname = self.dirname, show_plot=self.show_plt)
            plot_position_evol(self.rf_params, h5data, 
                               output_freq = self.dt_mon, 
                               style = self.lstyle, dirname = self.dirname, show_plot=self.show_plt)
            plot_energy_evol(self.rf_params, h5data, output_freq = self.dt_mon, 
                             style = self.lstyle, dirname = self.dirname,show_plot=self.show_plt)
            plot_COM_motion(self.general_params, self.rf_params, h5data,
                            output_freq = self.dt_mon, dirname = self.dirname, show_plot=self.show_plt)
            plot_transmitted_particles(self.rf_params, h5data, 
                                       output_freq = self.dt_mon, 
                                       style = self.lstyle, 
                                       dirname = self.dirname, show_plot=self.show_plt)
                    
            if self.PL:
                plot_PL_RF_freq(self.rf_params, h5data, 
                                output_freq = self.dt_mon,
                                dirname = self.dirname, show_plot=self.show_plt)
                plot_PL_RF_phase(self.rf_params, h5data, 
                                 output_freq = self.dt_mon,
                                 dirname = self.dirname, show_plot=self.show_plt)
                plot_PL_bunch_phase(self.rf_params, h5data, 
                                    output_freq = self.dt_mon, 
                                    dirname = self.dirname, show_plot=self.show_plt)
                plot_PL_phase_corr(self.rf_params, h5data, 
                                   output_freq = self.dt_mon, 
                                   dirname = self.dirname,show_plot=self.show_plt)
                plot_PL_freq_corr(self.rf_params, h5data, 
                                  output_freq = self.dt_mon, 
                                  dirname = self.dirname,show_plot=self.show_plt)
                plot_RF_phase_error(self.rf_params, h5data, 
                                    output_freq = self.dt_mon, 
                                    dirname = self.dirname,show_plot=self.show_plt)
                plot_RL_radial_error(self.rf_params, h5data, 
                                     output_freq = self.dt_mon, 
                                     dirname = self.dirname,show_plot=self.show_plt)
            
            if self.noiseFB:
                plot_LHCNoiseFB(self.rf_params, self.noiseFB, h5data, 
                                output_freq = self.dt_mon, 
                                dirname = self.dirname, show_plot=self.show_plt)
                plot_LHCNoiseFB_FWHM(self.rf_params, self.noiseFB, h5data, 
                                     output_freq = self.dt_mon, 
                                     dirname = self.dirname, show_plot=self.show_plt)
            
                if self.noiseFB.bl_meas_bbb != None:
                    plot_LHCNoiseFB_FWHM_bbb(self.rf_params, self.noiseFB, 
                                             h5data, output_freq = self.dt_mon, 
                                             dirname = self.dirname, show_plot=self.show_plt)

            h5data.close()

    def reset_frame(self, xmin, xmax, ymin, ymax):

        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax




# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
:Authors: **Danilo Quartullo, Konstantinos Iliakis**
'''

from __future__ import division

from builtins import range

import numpy as np
from scipy.constants import e

from ..utils import bmath as bm


class Music:

    r"""
    Implementation of the MuSiC algorithm in C++ to calculate the exact induced
    voltage generated by resonant modes in time domain without using slices,
    cost = O(n). The corresponding methods in Python are kept for reference.
    The method track_classic, which calculates in time domain the
    exact voltage with the O(n^2) algorithm used in the usual voltage
    definition, is kept just for reference.

    Parameters
    ----------
    Beam : object
        Beam object.
    resonator : float list
        List of the resonator parameters:
        [shunt impedance [:math:`\Omega`], angular resonant frequency [rad/s],
        quality factor [1]].
    n_macroparticles : int
        Number of macro-particles [1].
    n_particles : float
        Beam intensity [1].
    t_rev : float
        Revolution period [s]

    Attributes
    ----------
    beam : object
        Beam object.
    R_S : float
        shunt impedance [:math:`\Omega`]
    omega_R : float
        angular resonant frequency [rad/s]
    Q : float
        quality factor [1]
    n_macroparticles : int
        Number of macro-particles [1].
    n_particles : float
        Beam intensity [1].
    alpha : float
        Definition dependent on previously defined attributes.
    omega_bar : float
        Definition dependent on previously defined attributes.
    const : float
        Definition dependent on previously defined attributes.
    induced_voltage : float array
        Output induced voltage [V] (multiplied by -1 for BLonD conventions)
    coeff1 : float
        Definition dependent on previously defined attributes.
    coeff2 : float
        Definition dependent on previously defined attributes.
    coeff3 : float
        Definition dependent on previously defined attributes.
    coeff4 : float
        Definition dependent on previously defined attributes.
    input_first_component : float
        First component of vertical array in MuSiC algorithm
    input_second_component : float
        Second component of vertical array in MuSiC algorithm
    t_rev : float
        Revolution period [s]
    last_dt: float
        Last longitudinal coordinate of the beam [s]
    array_parameters : float array
        Array gathering four attributes already defined to be used in the C++
        algorithm.

    Notes
    -----
    The energies dE of the particles in the beam object are updated after the
    induced voltage calculation.

    See Also
    --------
    The MuSiC algorithm is described in:
    M. Migliorati, L. Palumbo, 'Multibunch and multiparticle simulation code
    with an alternative approach to wakefield effects', Phys. Rev. ST Accel.
    Beams 18, 2015.

    """

    def __init__(self, Beam, resonator, n_macroparticles, n_particles, t_rev):

        self.beam = Beam
        self.R_S = resonator[0]
        self.omega_R = resonator[1]
        self.Q = resonator[2]
        self.n_macroparticles = n_macroparticles
        self.n_particles = n_particles
        self.alpha = self.omega_R / (2 * self.Q)
        self.omega_bar = np.sqrt(self.omega_R ** 2 - self.alpha ** 2)
        self.const = -e * self.R_S * self.omega_R * \
            self.n_particles / (self.n_macroparticles * self.Q)
        self.induced_voltage = np.zeros(len(self.beam.dt))
        self.induced_voltage[0] = self.const / 2
        self.coeff1 = -self.alpha / self.omega_bar
        self.coeff2 = -self.R_S * self.omega_R / (self.Q * self.omega_bar)
        self.coeff3 = self.omega_R * self.Q / (self.R_S * self.omega_bar)
        self.coeff4 = self.alpha / self.omega_bar
        self.input_first_component = 1
        self.input_second_component = 0
        self.t_rev = t_rev
        self.last_dt = self.beam.dt[-1]
        self.array_parameters = np.array([self.input_first_component,
                                          self.input_second_component, self.t_rev, self.last_dt])

    def track_cpp(self):
        r"""
        Voltage in time domain (single-turn) using MuSiC (C++ code).
        Note: this method should also be called at turn number 1 when
        multi-turn voltage computations are needed.

        Examples
        --------
        >>> import impedances.music as musClass
        >>> from setup_cpp import libblond
        >>>
        >>> music_cpp = musClass.Music(my_beam, [R_S, 2*np.pi*frequency_R, Q],
        >>>                               n_macroparticles, n_particles, t_rev)
        >>> music_cpp.track_cpp()

        """
        bm.music_track(self.beam.dt, self.beam.dE, self.induced_voltage,
                       self.array_parameters, self.alpha, self.omega_bar,
                       self.const, self.coeff1, self.coeff2, self.coeff3,
                       self.coeff4)

    def track_cpp_multi_turn(self):
        r"""
        Voltage in time domain (multi-turn) using MuSiC (C++ code).
        Note: this method should be called from turn number 2 onwards when
        multi-turn voltage computations are needed..

        Examples
        --------
        >>> import impedances.music as musClass
        >>> from setup_cpp import libblond
        >>>
        >>> music_cpp = musClass.Music(my_beam, [R_S, 2*np.pi*frequency_R, Q],
        >>>                               n_macroparticles, n_particles, t_rev)
        >>> music_cpp.track_cpp()
        >>> for i in range(2, n_turns):
        >>>     music_cpp.track_cpp_multi_turn()

        """
        bm.music_track_multiturn(self.beam.dt, self.beam.dE, self.induced_voltage,
                                 self.array_parameters, self.alpha, self.omega_bar,
                                 self.const, self.coeff1, self.coeff2, self.coeff3,
                                 self.coeff4)

    def track_py(self):
        r"""
        Voltage in time domain (single-turn) using MuSiC (Python code).
        Note: this method should also be called at turn number 1 when
        multi-turn voltage computations are needed.

        Examples
        --------
        >>> import impedances.music as musClass
        >>>
        >>> music_cpp = musClass.Music(my_beam, [R_S, 2*np.pi*frequency_R, Q],
        >>>                               n_macroparticles, n_particles, t_rev)
        >>> music_cpp.track_py()

        """

        indices_sorted = np.argsort(self.beam.dt)
        self.beam.dt = self.beam.dt[indices_sorted]
        self.beam.dE = self.beam.dE[indices_sorted]
        self.beam.dE[0] += self.induced_voltage[0]
        self.input_first_component = 1
        self.input_second_component = 0

        for i in range(len(self.beam.dt) - 1):

            time_difference = self.beam.dt[i + 1] - self.beam.dt[i]

            exp_term = np.exp(-self.alpha * time_difference)
            cos_term = np.cos(self.omega_bar * time_difference)
            sin_term = np.sin(self.omega_bar * time_difference)

            product_first_component = exp_term * \
                ((cos_term + self.coeff1 * sin_term) * self.input_first_component
                 + self.coeff2 * sin_term * self.input_second_component)
            product_second_component = exp_term * \
                (self.coeff3 * sin_term * self.input_first_component
                 + (cos_term + self.coeff4 * sin_term) * self.input_second_component)

            self.induced_voltage[i + 1] = self.const * \
                (0.5 + product_first_component)
            self.beam.dE[i + 1] += self.induced_voltage[i + 1]

            self.input_first_component = product_first_component + 1.0
            self.input_second_component = product_second_component

        self.last_dt = self.beam.dt[-1]

    def track_py_multi_turn(self):
        r"""
        Voltage in time domain (multi-turn) using MuSiC (Python code).
        Note: this method should be called from turn number 2 onwards when
        multi-turn voltage computations are needed..

        Examples
        --------
        >>> import impedances.music as musClass
        >>>
        >>> music_cpp = musClass.Music(my_beam, [R_S, 2*np.pi*frequency_R, Q],
        >>>                               n_macroparticles, n_particles, t_rev)
        >>> music_cpp.track_py()
        >>> for i in range(2, n_turns):
        >>>     music_cpp.track_py_multi_turn()

        """

        indices_sorted = np.argsort(self.beam.dt)
        self.beam.dt = self.beam.dt[indices_sorted]
        self.beam.dE = self.beam.dE[indices_sorted]
        time_difference_0 = self.beam.dt[0] + self.t_rev - self.last_dt
        exp_term = np.exp(-self.alpha * time_difference_0)
        cos_term = np.cos(self.omega_bar * time_difference_0)
        sin_term = np.sin(self.omega_bar * time_difference_0)
        product_first_component = exp_term * \
            ((cos_term + self.coeff1 * sin_term) * self.input_first_component
             + self.coeff2 * sin_term * self.input_second_component)
        product_second_component = exp_term * \
            (self.coeff3 * sin_term * self.input_first_component
             + (cos_term + self.coeff4 * sin_term) * self.input_second_component)
        self.induced_voltage[0] = self.const * \
            (0.5 + product_first_component)
        self.beam.dE[0] += self.induced_voltage[0]
        self.input_first_component = product_first_component + 1.0
        self.input_second_component = product_second_component

        for i in range(len(self.beam.dt) - 1):

            time_difference = self.beam.dt[i + 1] - self.beam.dt[i]

            exp_term = np.exp(-self.alpha * time_difference)
            cos_term = np.cos(self.omega_bar * time_difference)
            sin_term = np.sin(self.omega_bar * time_difference)

            product_first_component = exp_term * \
                ((cos_term + self.coeff1 * sin_term) * self.input_first_component
                 + self.coeff2 * sin_term * self.input_second_component)
            product_second_component = exp_term * \
                (self.coeff3 * sin_term * self.input_first_component
                 + (cos_term + self.coeff4 * sin_term) * self.input_second_component)

            self.induced_voltage[i + 1] = self.const * \
                (0.5 + product_first_component)
            self.beam.dE[i + 1] += self.induced_voltage[i + 1]

            self.input_first_component = product_first_component + 1.0
            self.input_second_component = product_second_component

        self.last_dt = self.beam.dt[-1]

    def track_classic(self):
        r"""
        Voltage in time domain using the basic definition (Python code)

        """

        indices_sorted = np.argsort(self.beam.dt)
        self.beam.dt = self.beam.dt[indices_sorted]
        self.beam.dE = self.beam.dE[indices_sorted]
        self.beam.dE[0] += self.induced_voltage[0]
        self.induced_voltage[1:] = 0

        for i in range(len(self.beam.dt) - 1):

            for j in range(i + 1):

                time_difference = self.beam.dt[i + 1] - self.beam.dt[j]
                exp_term = np.exp(-self.alpha * time_difference)
                cos_term = np.cos(self.omega_bar * time_difference)
                sin_term = np.sin(self.omega_bar * time_difference)
                self.induced_voltage[i + 1] += \
                    exp_term * (cos_term + self.coeff1 * sin_term)

            self.induced_voltage[i + 1] = \
                self.const * (0.5 + self.induced_voltage[i + 1])
            self.beam.dE[i + 1] += self.induced_voltage[i + 1]




# Copyright 2015 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Methods to generate RF phase noise from noise spectrum and feedback noise
amplitude as a function of bunch length**

:Authors: **Helga Timko**
'''

from __future__ import division, print_function

from builtins import range

import numpy as np
import numpy.random as rnd

from ..plots.plot import fig_folder
from ..plots.plot_llrf import plot_phase_noise, plot_noise_spectrum
from ..toolbox.next_regular import next_regular

cfwhm = np.sqrt(2. / np.log(2.))


class FlatSpectrum:

    def __init__(self, Ring, RFStation, delta_f=1,
                 corr_time=10000, fmin_s0=0.8571, fmax_s0=1.1,
                 initial_amplitude=1.e-6, seed1=1234, seed2=7564,
                 predistortion=None, continuous_phase=False, folder_plots='fig_noise', print_option=True, initial_final_turns=[0, -1]):
        '''
        Generate phase noise from a band-limited spectrum.
        Input frequency band using 'fmin' and 'fmax' w.r.t. the synchrotron 
        frequency. Input double-sided spectrum amplitude [rad^2/Hz] using 
        'initial_amplitude'. Fix seeds to obtain reproducible phase noise.
        Select 'time_points' suitably to resolve the spectrum in frequency 
        domain. After 'corr_time' turns, the seed is changed to cut numerical
        correlated sequences of the random number generator.
        '''
        self.total_n_turns = Ring.n_turns
        self.initial_final_turns = initial_final_turns
        if self.initial_final_turns[1] == -1:
            self.initial_final_turns[1] = self.total_n_turns + 1

        self.f0 = Ring.f_rev[self.initial_final_turns[0]:self.initial_final_turns[1]]  # revolution frequency in Hz
        self.delta_f = delta_f           # frequency resolution [Hz]
        self.corr = corr_time           # adjust noise every 'corr' time steps
        self.fmin_s0 = fmin_s0                # spectrum lower bound in synchr. freq.
        self.fmax_s0 = fmax_s0                # spectrum upper bound in synchr. freq.
        self.A_i = initial_amplitude    # initial spectrum amplitude [rad^2/Hz]
        self.seed1 = seed1
        self.seed2 = seed2
        self.predistortion = predistortion
        if self.predistortion == 'weightfunction':
            # Overwrite frequencies
            self.fmin_s0 = 0.8571
            self.fmax_s0 = 1.001
        self.fs = RFStation.omega_s0[self.initial_final_turns[0]:self.initial_final_turns[1]] / (2 * np.pi)  # synchrotron frequency in Hz
        self.n_turns = len(self.fs) - 1
        self.dphi = np.zeros(self.n_turns + 1)
        self.continuous_phase = continuous_phase
        if self.continuous_phase:
            self.dphi2 = np.zeros(self.n_turns + 1 + self.corr / 4)
        self.folder_plots = folder_plots
        self.print_option = print_option

    def spectrum_to_phase_noise(self, freq, spectrum, transform=None):

        nf = len(spectrum)
        fmax = freq[nf - 1]

        # Resolution in time domain
        if transform is None or transform == 'r':
            nt = 2 * (nf - 1)
            dt = 1 / (2 * fmax)  # in [s]
        elif transform == 'c':
            nt = nf
            dt = 1. / fmax  # in [s]
        else:
            # NoiseError
            raise RuntimeError('ERROR: The choice of Fourier transform for the\
             RF noise generation could not be recognized. Use "r" or "c".')

        # Generate white noise in time domain
        rnd.seed(self.seed1)
        r1 = rnd.random_sample(nt)
        rnd.seed(self.seed2)
        r2 = rnd.random_sample(nt)
        if transform is None or transform == 'r':
            Gt = np.cos(2 * np.pi * r1) * np.sqrt(-2 * np.log(r2))
        elif transform == 'c':
            Gt = np.exp(2 * np.pi * 1j * r1) * np.sqrt(-2 * np.log(r2))

        # FFT to frequency domain
        if transform is None or transform == 'r':
            Gf = np.fft.rfft(Gt)
        elif transform == 'c':
            Gf = np.fft.fft(Gt)

        # Multiply by desired noise probability density
        if transform is None or transform == 'r':
            s = np.sqrt(2 * fmax * spectrum)  # in [rad]
        elif transform == 'c':
            s = np.sqrt(fmax * spectrum)  # in [rad]
        dPf = s * Gf.real + 1j * s * Gf.imag  # in [rad]

        # FFT back to time domain to get final phase shift
        if transform is None or transform == 'r':
            dPt = np.fft.irfft(dPf)  # in [rad]
        elif transform == 'c':
            dPt = np.fft.ifft(dPf)  # in [rad]

        # Use only real part for the phase shift and normalize
        self.t = np.linspace(0, float(nt * dt), nt)
        self.dphi_output = dPt.real

    def generate(self):

        for i in range(0, int(np.ceil(self.n_turns / self.corr))):

            # Scale amplitude to keep area (phase noise amplitude) constant
            k = i * self.corr       # current time step
            ampl = self.A_i * self.fs[0] / self.fs[k]

            # Calculate the frequency step
            f_max = self.f0[k] / 2
            n_points_pos_f_incl_zero = int(np.ceil(f_max / self.delta_f) + 1)
            nt = 2 * (n_points_pos_f_incl_zero - 1)
            nt_regular = next_regular(int(nt))
            if nt_regular % 2 != 0 or nt_regular < self.corr:
                # NoiseError
                raise RuntimeError('Error in noise generation!')
            n_points_pos_f_incl_zero = int(nt_regular / 2 + 1)
            freq = np.linspace(0, float(f_max), n_points_pos_f_incl_zero)
            delta_f = f_max / (n_points_pos_f_incl_zero - 1)

            # Construct spectrum
            nmin = int(np.floor(self.fmin_s0 * self.fs[k] / delta_f))
            nmax = int(np.ceil(self.fmax_s0 * self.fs[k] / delta_f))

            # To compensate the notch due to PL at central frequency
            if self.predistortion == 'exponential':

                spectrum = np.concatenate((np.zeros(nmin), ampl * np.exp(
                    np.log(100.) * np.arange(0, nmax - nmin + 1) / (nmax - nmin)),
                    np.zeros(n_points_pos_f_incl_zero - nmax - 1)))

            elif self.predistortion == 'linear':

                spectrum = np.concatenate((np.zeros(nmin),
                                           np.linspace(0, float(ampl), nmax - nmin + 1), np.zeros(n_points_pos_f_incl_zero - nmax - 1)))

            elif self.predistortion == 'hyperbolic':

                spectrum = np.concatenate((np.zeros(nmin),
                                           ampl * np.ones(nmax - nmin + 1) *
                                           1 / (1 + 0.99 * (nmin - np.arange(nmin, nmax + 1))
                                                / (nmax - nmin)), np.zeros(n_points_pos_f_incl_zero - nmax - 1)))

            elif self.predistortion == 'weightfunction':

                frel = freq[nmin:nmax + 1] / self.fs[k]  # frequency relative to fs0
                frel[np.where(frel > 0.999)[0]] = 0.999  # truncate center freqs
                sigma = 0.754  # rms bunch length in rad corresponding to 1.2 ns
                gamma = 0.577216
                weight = (4. * np.pi * frel / sigma**2)**2 * \
                    np.exp(-16. * (1. - frel) / sigma**2) + \
                    0.25 * (1 + 8. * frel / sigma**2 *
                            np.exp(-8. * (1. - frel) / sigma**2) *
                            (gamma + np.log(8. * (1. - frel) / sigma**2) +
                             8. * (1. - frel) / sigma**2))**2
                weight /= weight[0]  # normalise to have 1 at fmin
                spectrum = np.concatenate((np.zeros(nmin), ampl * weight,
                                           np.zeros(n_points_pos_f_incl_zero - nmax - 1)))

            else:
                spectrum = np.concatenate((np.zeros(nmin),
                                           ampl * np.ones(nmax - nmin + 1), np.zeros(n_points_pos_f_incl_zero - nmax - 1)))

            # Fill phase noise array
            if i < int(self.n_turns / self.corr) - 1:
                kmax = (i + 1) * self.corr
            else:
                kmax = self.n_turns + 1

            self.spectrum_to_phase_noise(freq, spectrum)
            self.seed1 += 239
            self.seed2 += 158
            self.dphi[k:kmax] = self.dphi_output[0:(kmax - k)]

            if self.continuous_phase:
                if i == 0:
                    self.spectrum_to_phase_noise(freq, spectrum)
                    self.seed1 += 239
                    self.seed2 += 158
                    self.dphi2[:self.corr / 4] = self.dphi_output[:self.corr / 4]

                self.spectrum_to_phase_noise(freq, spectrum)
                self.seed1 += 239
                self.seed2 += 158
                self.dphi2[(k + self.corr / 4):(kmax + self.corr / 4)] = self.dphi_output[0:(kmax - k)]

            if self.folder_plots is not None:
                fig_folder(self.folder_plots)
                plot_noise_spectrum(freq, spectrum, sampling=1, figno=i,
                                    dirname=self.folder_plots)
                plot_phase_noise(self.t[0:(kmax - k)], self.dphi_output[0:(kmax - k)],
                                 sampling=1, figno=i, dirname=self.folder_plots)

            rms_noise = np.std(self.dphi_output)
            if self.print_option:
                print("RF noise for time step %.4e s (iter %d) has r.m.s. phase %.4e rad (%.3e deg)"
                      % (self.t[1], i, rms_noise, rms_noise * 180 / np.pi))

        if self.continuous_phase:
            psi = np.arange(0, self.n_turns + 1) * 2 * np.pi / self.corr
            self.dphi = self.dphi * np.sin(psi[:self.n_turns + 1]) + self.dphi2[:(self.n_turns + 1)] * np.cos(psi[:self.n_turns + 1])

        if self.initial_final_turns[0] > 0 or self.initial_final_turns[1] < self.total_n_turns + 1:
            self.dphi = np.concatenate((np.zeros(self.initial_final_turns[0]), self.dphi, np.zeros(1 + self.total_n_turns - self.initial_final_turns[1])))


class LHCNoiseFB:
    '''
    *Feedback on phase noise amplitude for LHC controlled longitudinal emittance
    blow-up using noise injection through cavity controller or phase loop.
    The feedback compares the FWHM bunch length of the bunch to a target value 
    and scales the phase noise to keep the targeted value.
    Activate the feedback either by passing it in RFStation or in
    the PhaseLoop object.
    Update the noise amplitude scaling using track().
    Pass the bunch pattern (occupied bucket numbers from 0...h-1) in buckets 
    for multi-bunch simulations; the feedback uses the average bunch length.*
    '''

    def __init__(self, RFStation, Profile, bl_target, gain=0.1e9,
                 factor=0.93, update_frequency=22500, variable_gain=True,
                 bunch_pattern=None):

        #: | *Import RFStation*
        self.rf_params = RFStation

        #: | *Import Profile*
        self.profile = Profile

        #: | *Phase noise scaling factor. Initially 0.*
        self.x = 0.

        #: | *Target bunch length [s], 4-sigma value.*
        self.bl_targ = bl_target

        #: | *Measured bunch length [s], FWHM.*
        self.bl_meas = bl_target

        #: | *Feedback recursion scaling factor.*
        self.a = factor

        #: | *Update feedback every n_update turns.*
        self.n_update = update_frequency

        #: | *Switch to use constant or variable gain*
        self.variable_gain = variable_gain

        #: | *Feedback gain [1/s].*
        if self.variable_gain:
            self.g = gain * (self.rf_params.omega_s0[0] /
                             self.rf_params.omega_s0)**2
        else:
            self.g = gain * np.ones(self.rf_params.n_turns + 1)

        #: | *Bunch pattern for multi-bunch simulations*
        self.bunch_pattern = bunch_pattern

        #: | *Function dictionary to calculate FWHM bunch length*
        fwhm_functions = {'single': self.fwhm_single_bunch,
                          'multi': self.fwhm_multi_bunch}
        if self.bunch_pattern is None:
            self.fwhm = fwhm_functions['single']
            self.bl_meas_bbb = None
        else:
            self.bunch_pattern = np.ascontiguousarray(self.bunch_pattern)
            self.bl_meas_bbb = np.zeros(len(self.bunch_pattern))
            self.fwhm = fwhm_functions['multi']

    def track(self):
        '''
        *Calculate PhaseNoise Feedback scaling factor as a function of measured
        FWHM bunch length.*
        '''

        # Track only in certain turns
        if (self.rf_params.counter[0] % self.n_update) == 0:

            # Update bunch length, every x turns determined in main file
            self.fwhm()

            # Update noise amplitude-scaling factor
            self.x = self.a * self.x + self.g[self.rf_params.counter[0]] * \
                (self.bl_targ - self.bl_meas)

            # Limit to range [0,1]
            if self.x < 0:
                self.x = 0
            if self.x > 1:
                self.x = 1

    def fwhm_interpolation(self, index, half_height):

        time_resolution = self.profile.bin_centers[1] - self.profile.bin_centers[0]

        left = self.profile.bin_centers[index[0]] - (self.profile.n_macroparticles[index[0]] -
                                                     half_height) / (self.profile.n_macroparticles[index[0]] -
                                                                     self.profile.n_macroparticles[index[0] - 1]) * time_resolution

        right = self.profile.bin_centers[index[-1]] + (self.profile.n_macroparticles[index[-1]]
                                                       - half_height) / (self.profile.n_macroparticles[index[-1]] -
                                                                         self.profile.n_macroparticles[index[-1] + 1]) * time_resolution

        return cfwhm * (right - left)

    def fwhm_single_bunch(self):
        '''
        *Single-bunch FWHM bunch length calculation with interpolation.*
        '''

        half_height = np.max(self.profile.n_macroparticles) / 2.
        index = np.where(self.profile.n_macroparticles > half_height)[0]

        self.bl_meas = self.fwhm_interpolation(index, half_height)

    def fwhm_multi_bunch(self):
        '''
        *Multi-bunch FWHM bunch length calculation with interpolation.*
        '''

        # Find correct RF buckets
        phi_RF = self.rf_params.phi_RF[0, self.rf_params.counter[0]]
        omega_RF = self.rf_params.omega_RF[0, self.rf_params.counter[0]]
        bucket_min = (phi_RF + 2. * np.pi * self.bunch_pattern) / omega_RF
        bucket_max = bucket_min + 2. * np.pi / omega_RF

        # Bunch-by-bunch FWHM bunch length
        for i in range(len(self.bunch_pattern)):

            bind = np.where((self.profile.bin_centers - bucket_min[i]) *
                            (self.profile.bin_centers - bucket_max[i]) < 0)[0]
            hheight = np.max(self.profile.n_macroparticles[bind]) / 2.
            index = np.where(self.profile.n_macroparticles[bind] > hheight)[0]
            self.bl_meas_bbb[i] = self.fwhm_interpolation(bind[index], hheight)

        # Average FWHM bunch length
        self.bl_meas = np.mean(self.bl_meas_bbb)
